"""Processes for clustering data"""
from pyppl import Proc, Box
from . import params, rimport
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pDist2Feats():
	"""
	@name:
		pDist2Feats
	@description:
		Convert a distance matrix to coordinates, using multidimensional scaling.
	@input:
		`infile:file`: The distance matrix, could be a full distance matrix, a triangle matrix or a pair-wise distance file
			- full dist matrix (full):
			```
				s1	s2	s3
			s1	0	1	1
			s2	1	0	1
			s3	1	1	0
			```
			- triangle matrix (upper/lower), could be also lower triangle
			```
				s1	s2	s3
			s1	0	1	1
			s2		0	1
			s3			0
			```
			- pair-wise (pair): (assuming auto-pair-wise distance = 0, that is: `s1	s1	0`)
			```
			s1	s2	1
			s1	s3	1
			s2	s3	1
			```
			- Both rownames and header are required.
	@output:
		`outfile:file`: The output coordinate file
	@args:
		`informat`: The format of the input file: full, triangle or pair. Default: full
			- Could also be upper, lower, pair
		`k`:        How many dimensions? Default: 2 (R^2)
	"""
	pDist2Feats            = Proc(desc = 'Convert a distance matrix to coordinates, using multidimensional scaling.')
	pDist2Feats.input      = "infile:file"
	pDist2Feats.output     = "outfile:file:{{i.infile | fn}}.xyz"
	pDist2Feats.args.infmt = 'full'
	pDist2Feats.args.k     = 2
	pDist2Feats.lang       = params.Rscript.value
	pDist2Feats.script     = "file:scripts/cluster/pDist2Feats.r"
	return pDist2Feats

@procfactory
def _pFeats2Dist():
	"""
	@name:
		pFeats2Dist
	@description:
		Calculate the distance of pairs of instances
	@input:
		`infile:file`: The input file. Rows are instances, columns are features.
	@output:
		`outfile:file`: The output distance matrix
	@args:
		`inopts`: the options for input file
			- `cnames`: Whether input file has column names
			- `rnames`: Whether input file has row names
		`transfm`: Transform the results in follow way. Default: `scale`
			- `scale`: scale the distance to [0, 1]
			- `similarity`: 1 - scaled distance
			- `False`: don't do any transformation.
		`na` : Replace NAs with? Default: `0`
		`method`: Method used to calculate the distance. See available below. Default: `euclidean`
			- euclidean, manhattan, minkowski, chebyshev, sorensen, gower, soergel, kulczynski_d,
			- canberra, lorentzian, intersection, non-intersection, wavehedges, czekanowski, motyka,
			- kulczynski_s, tanimoto, ruzicka, inner_product, harmonic_mean, cosine, hassebrook,
			- jaccard, dice, fidelity, bhattacharyya, hellinger, matusita, squared_chord,
			- quared_euclidean, pearson, neyman, squared_chi, prob_symm, divergence, clark,
			- additive_symm, kullback-leibler, jeffreys, k_divergence, topsoe, jensen-shannon,
			- jensen_difference, taneja, kumar-johnson, avg
			- see: https://cran.r-project.org/web/packages/philentropy/vignettes/Distances.html
	@requires:
		`r-philentropy`
	"""
	pFeats2Dist              = Proc(desc = 'Calculate the distance between each pair of rows')
	pFeats2Dist.input        = 'infile:file'
	pFeats2Dist.output       = 'outfile:file:{{i.infile | fn}}.dist.txt'
	pFeats2Dist.args.inopts  = Box(cnames = True, rnames = True)
	pFeats2Dist.args.transfm = 'scale'
	pFeats2Dist.args.na      = 0
	pFeats2Dist.args.method  = 'euclidean'
	pFeats2Dist.envs.rimport = rimport
	pFeats2Dist.lang         = params.Rscript.str()
	pFeats2Dist.script       = "file:scripts/cluster/pFeats2Dist.r"
	return pFeats2Dist

@procfactory
def _pCluster():
	"""
	@name:
		pCluster
	@description:
		Use `optCluster` to select the best number of clusters and cluster method, then perform the clustering
	@input:
		`infile:file`: The input matrix file. Clustering will be performed against rows. If not, set `args.transpose` = True
	@output:
		`outfile:file`: The output cluster file
		`outdir:dir`:   The output directory containing the figures
	@args:
		`transpose`:    Transpose the input matrix. Default: False
		`cnames`:       Whether the input matrix contains header before transposing. Default: False
		`rnames`:       Which column is the rownames before transposing. Default: 1
		`plot`:         Whether plot the cluster. Default: True
		`minc`:         Min number of clusters to test. Default: 2
		`maxc`:         Min number of clusters to test. Default: 15
			- If number of rows (nrows) <= 15, then max = nrows - 1
		`methods`:      The methods to test. Default: "all"
			- Could be any of "agnes", "clara", "diana", "fanny", "hierarchical", "kmeans", "model", "pam", "som", "sota", "em.nbinom", "da.nbinom", "sa.nbinom", "em.poisson", "da.poisson", "sa.poisson"
			- Multiple methods could be separated by comma (,), or put in a list
			- By default, fanny, model and sota will be excluded because fanny causes error and the latter two are slow. You can manually include them if you want.
			- Improper methods will be automatically excluded by `args.isCount`
		`isCount`:      Whether the data is count data. Corresponding methods will be tested. Default: False
	@requires:
		[`r-optCluster`](https://rdrr.io/cran/optCluster/man/optCluster.html)
		[`r-factoextra`](https://cran.r-project.org/web/packages/factoextra/index.html)
	"""
	pCluster                = Proc (desc = 'Use `optCluster` to select the best number of clusters and cluster method, then perform the clustering.')
	pCluster.input          = 'infile:file'
	pCluster.output         = 'outfile:file:{{i.infile | fn}}.cluster/clusters.txt, outdir:dir:{{i.infile | fn}}.cluster'
	pCluster.args.transpose = False
	pCluster.args.cnames    = True
	pCluster.args.rnames    = True
	pCluster.args.plot      = True
	pCluster.args.minc      = 2
	pCluster.args.maxc      = 15
	# "agnes", "clara", "diana", "fanny", "hierarchical", "kmeans", "model", "pam", "som", "sota", "em.nbinom", "da.nbinom", "sa.nbinom", "em.poisson", "da.poisson", "sa.poisson"
	# Note fanny reports error and the algorithm cannot go further.
	# It is excluded by default, you can add it manually
	# model and sota are excluded because they are slow
	# You can also manually add them
	pCluster.args.methods   = 'all'
	pCluster.args.isCount   = False
	pCluster.args.devpars   = Box(res=300, width=2000, height=2000)
	pCluster.lang           = params.Rscript.value
	pCluster.script         = "file:scripts/cluster/pCluster.r"
	return pCluster


@procfactory
def _pMCluster():
	"""
	@name:
		pMCluster
	@description:
		Use `r-mclust` to do clustering. Current just do simple clustering with the package
	@input:
		`infile:file`: The input a coordinate file
	@output:
		`outdir:dir`:  The output of final results
	@args:
		`transpose`: Transpose the input matrix? Default: False
		`rnames`:  The `row.names` for `read.table` to read the input file, default: True.
		`cnames`:  The `header` argument for `read.table` to read the input file, default: True.
		`caption`: The caption for the `fviz_cluster`, default: "CLARA Clustering".
		`minc`:    The min # clusters to try, default: 2
		`maxc`:    The max # clusters to try, default: 15
	@requires:
		[`r-mclust`](https://cran.r-project.org/web/packages/mclust/index.html)
		[`r-factoextra`](https://cran.r-project.org/web/packages/factoextra/index.html)
	"""
	pMCluster                = Proc(desc = 'Use `r-mclust` to do clustering. Current just do simple clustering with the package.')
	pMCluster.input          = "infile:file"
	pMCluster.output         = 'outfile:file:{{i.infile | fn}}.mclust/clusters.txt, outdir:dir:{{i.infile | fn}}.mclust'
	pMCluster.args.transpose = False
	pMCluster.args.rnames    = True
	pMCluster.args.cnames    = True
	pMCluster.args.minc      = 2
	pMCluster.args.maxc      = 15
	pMCluster.args.devpars   = Box(res=300, width=2000, height=2000)
	pMCluster.lang           = params.Rscript.value
	pMCluster.script         = "file:scripts/cluster/pMCluster.r"
	return pMCluster


@procfactory
def _pAPCluster():
	"""
	@name:
		pAPCluster
	@description:
		Use `r-apcluster` to do clustering.
	@input:
		`infile:file`:  The input a coordinate file
	@output:
		`outdir:dir`:   The output of final results
	@args:
		`transpose`: Transpose the input matrix? Default: False
		`rownames`:     The `row.names` for `read.table` to read the input file, default: 1.
		`header`:       The `header` argument for `read.table` to read the input file, default: True.
		`caption`:      The caption for the `fviz_cluster`, default: "APClustering".
	@requires:
		[`r-apcluster`](https://cran.r-project.org/web/packages/apcluster/index.html)
		[`r-factoextra`](https://cran.r-project.org/web/packages/factoextra/index.html)
	"""
	pAPCluster                = Proc(desc = 'Use `r-apcluster` to do clustering.')
	pAPCluster.input          = "infile:file"
	pAPCluster.output         = 'outfile:file:{{i.infile | fn}}.apclust/clusters.txt, outdir:dir:{{i.infile | fn}}.apclust'
	pAPCluster.args.transpose = False
	pAPCluster.args.rnames    = True
	pAPCluster.args.cnames    = True
	pAPCluster.args.devpars   = Box(res=300, width=2000, height=2000)
	pAPCluster.lang           = params.Rscript.value
	pAPCluster.script         = "file:scripts/cluster/pAPCluster.r"
	return pAPCluster

@procfactory
def _pHCluster():
	"""
	@name:
		pHCluster
	@description:
		Do hierarchical clustering.
	@input:
		`infile:file`: The input files with variants as rows, features as columns.
			- NOTE: clustering is performed on rows, rownames are the leaf labels.
	@output:
		`outdir:dir`:  The result directory, containing:
			- `hclust.merge.txt`: including merge and height information
			- `hclust.order.txt`: including order and labels information
			- `hclust.png`:       the dendrogram plot
	@args:
		`fast`:     whether to use `fastcluster` package or not, default: False
		`gg`:       whether to use `ggdendro` or not, default: False
		`rownames`: The `row.names` for `read.table` to read the input file, default: 1.
		`header`:   The `header` argument for `read.table` to read the input file, default: True.
		`method`:   Which method to use for `hclust`. Default: "complete" (use `?hclust` to check all availables)
		`rotate`:   Which to rotate the plot or not. Default: False
		`transpose`:Whether to transpose the matrix before cluster. Default: False
	@requires:
		[`r-fastcluster`](https://cran.r-project.org/web/packages/fastcluster/index.html) if `args.fast` is True
		[`r-ggdendro`](https://cran.r-project.org/web/packages/ggdendro/index.html) if `args.gg` is True
	"""
	pHCluster = Proc(desc = 'Do hierarchical clustering.')
	pHCluster.input  = "infile:file"
	pHCluster.output = [
		"outmerge:file:{{i.infile | fn}}.hclust/merge.txt",
		"outorder:file:{{i.infile | fn}}.hclust/order.txt",
		"outdir:dir:{{i.infile | fn}}.hclust"
	]
	pHCluster.args.transpose = False
	pHCluster.args.cnames    = True
	pHCluster.args.rnames    = True
	pHCluster.args.method    = 'complete'
	pHCluster.args.rotate    = False
	pHCluster.args.fast      = False
	pHCluster.args.devpars   = Box(res=300, width=2000, height=2000)
	pHCluster.lang           = params.Rscript.value
	pHCluster.script         = "file:scripts/cluster/pHCluster.r"
	return pHCluster

@procfactory
def _pKMeans():
	"""
	@name:
		pKMeans
	"""
	pKMeans        = Proc(desc = 'Do K-Means clustering')
	pKMeans.input  = 'infile:file'
	pKMeans.output = [
		'outfile:file:{{i.infile | fn}}.kmeans/{{i.infile | fn}}.kmeans.txt',
		'outdir:dir:{{i.infile | fn}}.kmeans'
	]
	pKMeans.args.inopts  = Box(cnames = True, rnames = True)
	pKMeans.args.k       = None
	pKMeans.args.plot    = True
	pKMeans.args.ggs     = Box()
	pKMeans.args.devpars = Box(res=300, width=2000, height=2000)
	pKMeans.envs.rimport = rimport
	pKMeans.lang         = params.Rscript.str()
	pKMeans.script       = "file:scripts/cluster/pKMeans.r"
	return pKMeans




from os import path
from enrichr import Enrichr
from pyppl import Box
from bioprocs import rimport
from bioprocs.utils import shell2 as shell
from bioprocs.utils.parallel import Parallel
from bioprocs.utils.gene import genenorm
from bioprocs.utils.tsvio2 import TsvReader

{% from pyppl.utils import alwaysList %}
infile   = {{i.infile | quote}}
prefix   = {{i.infile | fn2 | quote}}
outdir   = {{o.outdir | quote}}
inopts   = {{args.inopts | repr}}
genecol  = {{args.genecol | ?bool | :_ | :0 | repr}}
top      = {{args.top}}
dbs      = {{args.libs | alwaysList | repr}}
plot     = {{args.plot | repr}}
nthread  = {{args.nthread | repr}}
Rscript  = {{args.Rscript | repr}}
cutoff   = {{args.cutoff | repr}}
devpars  = {{args.devpars | repr}}
pathview = {{args.pathview | repr}}
include  = {{args.include | ?bool | :_ | :None }}

shell.load_config(Rscript = Rscript)

if isinstance(cutoff, dict):
	if cutoff['by'] == 'p':
		cutoff['by'] = 'Pval'
	if cutoff['by'] == 'q':
		cutoff['by'] = 'AdjPval'

reader = TsvReader(infile, **inopts)
genes  = list(set([r[genecol] for r in reader if not include or include(r)]))
with open(path.join(outdir, 'genes.txt'), 'w') as fg:
	fg.write('\n'.join(genes))

en = Enrichr(cutoff = cutoff, top = top, Rscript = Rscript)
en.addList(genes, description = path.basename(infile))

para = Parallel(nthread = nthread)
runPathview = lambda r, hsa: shell.fg.Rscript(r, hsa)
for db in dbs:
	outfile = path.join(outdir, prefix + '.' + db + '.txt')
	en.enrich(db)
	en.export(outfile, top = 100)
	if plot:
		plotfile = path.join(outdir, prefix + '.' + db + '.png')
		en.plot(plotfile, res = devpars.res, width = devpars.width, height = devpars.height)
	if pathview and 'KEGG' in db:
		pathviewRDir  = path.join(outdir, prefix + '.' + db + '.pathview')
		pathviewRfile = path.join(pathviewRDir, 'pathview.R')
		shell.mkdir(pathviewRDir)
		with open(pathviewRfile, 'w') as f:
			f.write("""
			{rimport}('__init__.r')
			library(pathview)
			args = commandArgs(trailingOnly = TRUE)
			setwd({pathviewRDir!r})
			inopts = {{args.inopts | R}}
			inopts$rnames = FALSE
			indata = read.table.inopts({infile!r}, inopts)
			genes  = as.vector(indata[, {genecol}, drop = TRUE])
			pvargs = {{args.pathview | R}}
			{% raw %}
			if (!is.null(pvargs$fccol)) {{
				fcdata = as.vector(indata[, pvargs$fccol, drop = TRUE])
				names(fcdata) = genes
				genes = fcdata
			}}
			{% endraw %}
			pathview(gene.data = genes, pathway.id = args[1], species = 'hsa', gene.idtype="SYMBOL")
			""".format(
				rimport = rimport, genecol = genecol + 1 if isinstance(genecol, int) else genecol,
				infile = infile, pathviewRDir = pathviewRDir)
			)
		para.run(runPathview, [(pathviewRfile, term.Term.split('_')[-1]) for term in en.results[:top]])




from os import path
from pyppl import Box
from bioprocs.utils import shell2 as shell

cnvkit   = {{args.cnvkit | quote}}
infile   = {{i.infile | quote}}
outfile  = {{o.outfile | quote}}
antifile = {{o.antifile | quote}}
target   = {{i.tgfile | quote}}
atarget  = {{i.atgfile | quote}}
nthread  = {{args.nthread | repr}}
params   = {{args.params | repr}}

shell.load_config(cnvkit = dict(
	_exe = cnvkit,
	_env = dict(
		OPENBLAS_NUM_THREADS = str(nthread),
		OMP_NUM_THREADS      = str(nthread),
		NUMEXPR_NUM_THREADS  = str(nthread),
		MKL_NUM_THREADS      = str(nthread)
	),
	_cwd = path.dirname(outfile)
))

params.p  = nthread
params1   = params.copy()
params2   = params.copy()
params1.o = outfile
params2.o = antifile

shell.fg.cnvkit.coverage(infile, target,  **params1)
shell.fg.cnvkit.coverage(infile, atarget, **params2)




from os import path
from pyppl import Box
from bioprocs.utils import shell2 as shell
from bioprocs.utils.reference import tabixIndex
from tempfile import gettempdir

infile   = {{i.infile | quote}}
outfile  = {{o.outfile | quote}}
params   = {{args.params | repr}}
nthread  = {{args.nthread | repr}}
bcftools = {{args.bcftools | quote}}
annfile  = {{args.annfile | quote}}
tabix    = {{args.tabix | quote}}
header   = {{args.header | repr}}
cols     = {{args.cols | repr}}

shell.load_config(bcftools = bcftools)

if annfile and path.isfile(annfile):
	abname   = path.basename(annfile)
	ext      = path.splitext(abname[:-3] if abname.endswith('.gz') else abname)[-1][1:]
	infile   = tabixIndex(infile, 'vcf', tabix)
	params.a = tabixIndex(annfile, ext, tabix)
if cols:
	params.c = ','.join(cols if isinstance(cols, list) else [c for c in cols.split(',')])
if header:
	if not isinstance(header, list):
		header = [header]
	from datetime import datetime
	now = str(datetime.now().timestamp())
	headerfile = path.join(gettempdir(), 'pBcftools.annotate.header.%s' % now)
	with open(headerfile, 'w') as fhead:
		for head in header:
			fhead.write(head + '\n')
	params.h = headerfile

params._ = infile 
params.o = outfile
params.threads = nthread

shell.fg.bcftools.annotate(**params)



from pyparam import commands

commands._prefix       = '-'

# command: params
commands.params            = 'Show or query a parameter'
commands.params._.type     = list
commands.params._.required = True
commands.params._.desc     = 'The name of the parameter.'
commands.params.get        = False
commands.params.get.desc   = 'Get the value.'

# command: list
commands.list          = 'List modules or process'
commands.list._hbald   = False
commands.list.all      = False
commands.list.a        = commands.list.all
commands.list.all.desc = 'List all processes if no module specified.'
commands.list._        = []
commands.list._.desc   = 'Module name.'

# command: proc
commands.proc = 'Search modules or processes.'
commands.proc._        = []
commands.proc._.desc   = 'Part of the module or process name.'

# command: completion
commands.completion = 'Generate completions for bioprocs.'
commands.completion.shell = 'auto'
commands.completion.shell.desc = 'The shell'
commands.completion.auto = False
commands.completion.auto.desc = 'Whether automatically deploy the completion script.'
commands.completion.s = commands.completion.shell
commands.completion.a = commands.completion.auto
commands.completions  = commands.completion

# command: profile
commands.profile        = 'List available running profiles.'
commands.profile._hbald = False
commands.profile._.desc = 'Additional configurations.'

def helpx(helps):
	from bioprocs.console.utils import Pipeline
	#helps.add('Parameter/Process query commands', helps.select('Available commands'))
	helps.add('Assembled pipelines', sectype = 'option', prefix = '')
	helps.add('Processes', sectype = 'option', prefix = '')
	helps.add('Other commands', sectype = 'option', prefix = '')
	help_avail    = helps.select('Available')
	help_pipeline = helps.select('Assembled')
	help_process  = helps.select('Processes')
	help_other    = helps.select('Other')

	for ppl in Pipeline.pipelines():
		help_pipeline.add((ppl, '', Pipeline(ppl).desc))

	help_process.add(help_avail.select('list'))
	help_process.add(help_avail.select('proc'))
	help_process.add(('<module>', '', [
		'List the process of module.', 'Same as "bioprocs list <module>"']))
	help_process.add(('<module.proc>', '', 'Run a process in command line.'))
	help_other.add(help_avail.select('completion'))
	help_other.add(help_avail.select('params'))
	help_other.add(help_avail.select('profile'))
	help_other.add(help_avail.select('help'))

	helps.remove('Available')


# modify help page
commands._helpx = helpx



"""A set of immunotherapy-related bioinformatics tools"""
from pyppl import Proc, Box
from . import params, delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pTopiary():
	"""
	@description:
		Predict mutation-derived cancer T-cell epitopes from somatic variants, tumor RNA expression data, and patient HLA type.
	@input:
		infile: The input VCF or MAF file.
		mhcallele: The MHC alleles in format of "HLA-A02:01,HLA-B07:02" or a file with one per line.
			- You can have star in them: "HLA-A*02:01,HLA-B*07:02"
		gexpr: Cufflinks FPKM tracking file containing gene expression estimates.
		texpr: Cufflinks FPKM tracking file containing transcript expression estimates.
	@output:
		outfile: The output csv file.
			- A html file will also be generated with suffix '.html'
	@args:
		topiary       (str): Path to topiary.
		netmhc        (str): Path to netmhc.
		netmhcpan     (str): Path to netmhcpan.
		netmhciipan   (str): Path to netmhciipan.
		netmhccons    (str): Path to netmhccons.
		smm           (str): Path to smm.
		smm_pmbec     (str): Path to smm-pmbec.
		mhc_predictor (str): The MHC binding predictor. Could be one of:
			- netmhc: Local NetMHC predictor (Topiary will attempt to automatically detect whether NetMHC 3.x or 4.0 is available)
			- netmhcpan: Local NetMHCpan predictor
			- netmhciipan: Local NetMHCIIpan predictor
			- netmhccons: Local NetMHCcons
			- random: Random IC50 values
			- smm: Local SMM predictor
			- smm-pmbec: Local SMM-PMBEC predictor
			- netmhcpan-iedb: Use NetMHCpan via the IEDB web API
			- netmhccons-iedb: Use NetMHCcons via the IEDB web API
			- smm-iedb: Use SMM via the IEDB web API
			- smm-pmbec-iedb: Use SMM-PMBEC via the IEDB web API
		params (Box): Other parameters for topiary
		wildtype (bool): Also output wildtype peptide binding affinity.
			- Only available for local MHC binding affinity predictor
		tmpdir (str): Temporary directory for running local MHC predictors
	"""
	return Box(
		desc   = 'Epitope prediction using Topiary',
		lang   = params.python.value,
		input  = 'infile:file, mhcallele:var, gexpr:file, texpr:file',
		output = 'outfile:file:{{i.infile | stem}}.txt',
		args   = Box(
			topiary       = params.topiary.value,
			netmhc        = params.netmhc.value,
			netmhcpan     = params.netmhcpan.value,
			netmhciipan   = params.netmhciipan.value,
			netmhccons    = params.netmhccons.value,
			genome        = params.genome.value,
			smm           = params.smm.value,
			smm_pmbec     = params.smm_pmbec.value,
			tmpdir        = params.tmpdir.value,
			mhc_predictor = 'netmhcpan',
			wildtype      = False,
			params        = Box({
				'output-csv-sep': '\t',

			})
		)
	)


"""TCGA data analysis"""
from pyppl import Proc, Box
from . import params, rimport
from .utils import fs2name
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pTCGADownload():
	"""
	@name:
		pTCGADownload
	@description:
		Download TCGA use `gdc-client` and a manifest file
	@input:
		`manifile:file`: the manifest file
	@output:
		`outdir:file`:   the directory containing downloaded file
	@args:
		`params`    : other params for `gdc-client download`, default: `{'no-file-md5sum': True}`
		`gdc_client`: the executable file of `gdc-client`,    default: "gdc-client"
		`nthread`   : Number of threads to use. Default     : `1`
		`token`     : The token file if needed.
	"""
	pTCGADownload                 = Proc (desc = 'Download data with gdc-client and a menifest file.')
	pTCGADownload.input           = "manifile:file"
	pTCGADownload.output          = "outdir:dir:{{i.manifile | fn}}"
	pTCGADownload.echo            = Box(jobs = 0, type = 'stderr')
	pTCGADownload.args.params     = Box({'no-file-md5sum': True})
	pTCGADownload.args.nthread    = 1
	pTCGADownload.args.token      = None
	pTCGADownload.args.gdc_client = params.gdc_client.value
	pTCGADownload.lang            = params.python.value
	pTCGADownload.script          = "file:scripts/tcga/pTCGADownload.py"
	return pTCGADownload

@procfactory
def _pSample2SubmitterID():
	"""
	@name:
		pSample2SubmitterID
	@description:
		convert TCGA sample names with submitter id with metadata and sample containing folder
	@input:
		`indir:file`:    the directory containing the samples
		`mdfile:file`: the metadata file
	@output:
		`outdir:file`: the directory containing submitter-id named files
	@args:
		`method`: How the deal with the files. Default: `symlink`
			- We can also do `copy`
		`nthread`: Number threads to use. Default: `1`
	"""
	pSample2SubmitterID              = Proc(desc = 'convert TCGA sample names with submitter id with metadata and sample containing folder')
	pSample2SubmitterID.input        = "indir:file, mdfile:file"
	pSample2SubmitterID.output       = "outdir:dir:{{i.indir | fn}}"
	pSample2SubmitterID.args.method  = 'symlink' # or copy
	pSample2SubmitterID.args.nthread = 1
	pSample2SubmitterID.lang         = params.python.value
	pSample2SubmitterID.script       = "file:scripts/tcga/pSample2SubmitterID.py"
	return pSample2SubmitterID

@procfactory
def _pGtFiles2Mat():
	"""
	@name:
		pGtFiles2Mat
	@description:
		Convert TCGA genotype files to a matrix.
	@input:
		`infiles:files`: The input genotypes files
	@output:
		`outfile:file`: The output matrix file
	@args:
		`rsmap`    : The rsid probe mapping file. If not provided, will use the probe id for matrix rownames.
		`fn2sample`: How to convert filename(without extension) to sample name. Default: `None`
		`confcut`  : The confidence cutoff. Genotype will be NA for lower confidence snps. Default: `0.05`
	"""
	pGtFiles2Mat                = Proc(desc = 'Convert genotype files to a matrix.')
	pGtFiles2Mat.input          = 'infiles:files'
	pGtFiles2Mat.output         = 'outfile:file:{{i.infiles | fs2name}}.gt.txt'
	pGtFiles2Mat.args.rsmap     = params.rsmap_gwsnp6.value
	pGtFiles2Mat.args.confcut   = 0.05
	pGtFiles2Mat.args.fn2sample = "lambda fn: fn.split('.')[0]"
	pGtFiles2Mat.envs.fs2name   = fs2name
	pGtFiles2Mat.lang           = params.python.value
	pGtFiles2Mat.script         = "file:scripts/tcga/pGtFiles2Mat.py"
	return pGtFiles2Mat

@procfactory
def _pClinic2Survival():
	"""
	@name:
		pClinic2Survival
	@description:
		Convert TCGA clinic data to survival data
		The clinic data should be downloaded as "BCR Biotab" format
	@input:
		`infile:file`: The clinic data file downloaded from TCGA
	@output:
		`outfile:file`: The output file
		`covfile:file`: The covariate file
	@args:
		`cols`: The column names:
			- `time_lastfollow`: The column names of last follow up. Default: `['days_to_last_followup']`
			- `time_death`: The column names of time to death. Default: `['days_to_death']`
			- `status`: The columns of vital status. Default: `['vital_status']`
			- `age`: The columns of days to birth. Default: `['days_to_birth']`
		`covs`: The covariates to output. Default:
			- `gender`, `race`, `ethnicity`, `age`
		`mat`: An expression or genotype matrix with samples as column names, used to get sample names for patient instead of short ones. Default: `None`
	"""
	pClinic2Survival           = Proc(desc = 'Convert TCGA clinic data to survival data.')
	pClinic2Survival.input     = 'infile:file'
	pClinic2Survival.output    = 'outfile:file:{{i.infile | stem}}.survdata.txt, covfile:file:{{i.infile | stem}}.survcov.txt'
	pClinic2Survival.args.cols = Box(
		time_lastfollow = ['days_to_last_followup'],
		time_death      = ['days_to_death'],
		status          = ['vital_status'],
		age             = ['days_to_birth'],
		patient         = ['bcr_patient_barcode']
	)
	pClinic2Survival.args.covs = ['gender', 'race', 'ethnicity', 'age']
	pClinic2Survival.args.mat  = None
	pClinic2Survival.lang      = params.python.value
	pClinic2Survival.script    = "file:scripts/tcga/pClinic2Survival.py"
	return pClinic2Survival

@procfactory
def _pClinic2PlinkMeta():
	"""
	@name:
		pClinic2PlinkMeta
	@description:
		Convert TCGA clinic data to plink meta file.
		Used with a `GTMat` file to convert to plink binary file (see vcfnext.pGTMat2Plink).
	@input:
		`infile:file`: The input TCGA patiant clinic file.
	@output:
		`outfile:file`: The meta file.
	@args:
		`suffix`: The suffix added to the barcode. Default: `''`
			- By default the barcode is like `TCGA-55-8087`, but in analysis, we want to be specific to samples
			- For example, for solid tumor, it'll be `TCGA-55-8087-01`, then suffix `-01` will be added
			- You can add multiple suffices by `-01, -10` or `['-01', '-10']`
	"""
	pClinic2PlinkMeta             = Proc(desc = 'Convert TCGA clinic data to plink meta file.')
	pClinic2PlinkMeta.input       = 'infile:file'
	pClinic2PlinkMeta.output      = 'outfile:file:{{i.infile | fn}}.meta.txt'
	pClinic2PlinkMeta.args.suffix = ''
	pClinic2PlinkMeta.lang        = params.python.value
	pClinic2PlinkMeta.script      = "file:scripts/tcga/pClinic2PlinkMeta.py"
	return pClinic2PlinkMeta

@procfactory
def _pClinic2Cov():
	"""
	@name:
		pClinic2Cov
	"""
	pClinic2Cov              = Proc(desc = 'Extract information from clinic data as covariance matrix')
	pClinic2Cov.input        = 'infile:file'
	pClinic2Cov.output       = 'outfile:file:{{i.infile | fn}}.cov.txt'
	pClinic2Cov.args.covs    = ['gender', 'race', 'pathologic_stage', 'age_at_initial_pathologic_diagnosis']
	pClinic2Cov.args.asnum   = True
	pClinic2Cov.args.suffix  = ''
	pClinic2Cov.envs.rimport = rimport
	pClinic2Cov.lang         = params.Rscript.value
	pClinic2Cov.script       = "file:scripts/tcga/pClinic2Cov.r"
	return pClinic2Cov




import cmdy
from pysam import VariantFile
from pyppl import Box
from os import path, environ
from bioprocs.utils import logger, shell2 as shell
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter, TsvRecord

inmuts   = {{i.muts | quote}}
incnvs   = {{i.cnvs | quote}}
outdir   = {{o.outdir | quote}}
pyclone  = {{args.pyclone | quote}}
bcftools = {{args.bcftools | quote}}
bedtools = {{args.bedtools | quote}}
params   = {{args.params | repr}}
nthread  = {{args.nthread | int }}
refgene  = {{args.refgene | quote }}
mutctrl  = {{args.mutctrl | repr}}
cnctrl   = {{args.cnctrl | repr}}

env = environ.copy()
env.update(dict(
	OPENBLAS_NUM_THREADS = str(nthread),
	OMP_NUM_THREADS      = str(nthread),
	NUMEXPR_NUM_THREADS  = str(nthread),
	MKL_NUM_THREADS      = str(nthread)
) if nthread else {})

shell.load_config(pyclone = dict(
	_exe = pyclone,
	_cwd = outdir,
	_env = env
), bcftools = bcftools, bedtools = bedtools)

def defaultCNs(chrom, gt):
	# normal_cn, minor_cn, major_cn
	ret = [2, 0, 2]
	if gt == 'AA':
		ret[0] = 1 if chrom == 'chrY' else 2 # sex information needed for chrX
		ret[1] = 0
		ret[2] = ret[0]
	elif gt == 'AB':
		ret[0] = 2
		ret[1] = 1
		ret[2] = 1
	else: # BB
		ret[0] = 1 if chrom == 'chrY' else 2
		ret[1] = ret[0]
		ret[2] = 0
	return ret

def singleSampleMutVCF2BED(vcffiles):
	"""
	Convert single(paired)-sample VCFs to a BED file like:
	#CHROM  START   END	    NAME      GENOTYPE REFCOUNTS VARCOUNTS CASE    VARFREQ
	chr1    70820   70820   chr:70820 BB       1083      1996      NA12156 0.648
	"""
	ret = {}
	for vcffile in vcffiles:
		samples = shell.bcftools.query(l = vcffile).strip().splitlines()
		assert len(samples) <= 2
		samidx  = 0 if len(samples) == 1 or mutctrl in (1, -1, None) else 1
		sample  = samples[samidx]
		ret[sample] = path.join(outdir, sample + '.muts.bed')
		writer = TsvWriter(ret[sample])
		writer.cnames = ['CHROM', 'START', 'END', 'NAME', 'GENOTYPE', 'REFCOUNTS', 'VARCOUNTS', 'CASE', 'VARFREQ']
		for line in shell.bcftools.query(
			_ = vcffile,
			s = sample,
			f = '%CHROM\t%POS\t%POS\t%CHROM:%POS\t%GT\t%FORMAT/AD{0}\t%FORMAT/AD{1}\t%SAMPLE\t%FORMAT/AF\n'):

			items = line.split('\t')
			if items[4] in ('0/0', '0|0'):
				items[4] = 'AA'
			elif items[4] in ('0/1', '0|1'):
				items[4] = 'AB'
			elif items[4] in ('1/1', '1|1'):
				items[4] = 'BB'
			writer.write(items)

		writer.close()
	return ret

def multiSampleMutVCF2BED(vcffile):
	ret = {}
	writers = {}
	samples = shell.bcftools.query(l = vcffile).strip().splitlines()
	samples = [sample for i, sample in samples
		if mutctrl is None or i != (mutctrl if mutctrl >= 0 else len(samples) + mutctrl)]
	for sample in samples:
		ret[sample] = path.join(outdir, sample + '.muts.bed')
		writer = TsvWriter(sample)
		writer.cnames = ['CHROM', 'START', 'END', 'NAME', 'GENOTYPE', 'REFCOUNTS', 'VARCOUNTS', 'CASE', 'VARFREQ']
		writers[sample] = writer
	for line in shell.bcftools.query(
		_ = vcffile,
		s = ','.join(samples),
		f = '[%CHROM\t%POS\t%POS\t%CHROM:%POS\t%GT\t%FORMAT/AD{0}\t%FORMAT/AD{1}\t%SAMPLE\t%FORMAT/AF\n]'):
		items = line.split('\t')

		writer = writers[items[7]]
		if items[4] in ('0/0', '0|0'):
			items[4] = 'AA'
		elif items[4] in ('0/1', '0|1'):
			items[4] = 'AB'
		elif items[4] in ('1/1', '1|1'):
			items[4] = 'BB'
		writer.write(items)

	for writer in writers.values():
		writer.close()
	return ret

def MAF2BED(maffile):

	reader = TsvReader(maffile)
	if 't_alt_count' not in reader.cnames:
		raise ValueError('t_alt_count not found in MAF file.')
	if 't_ref_count' not in reader.cnames:
		raise ValueError('t_ref_count not found in MAF file.')

	ret = {}
	writers = {}
	for r in reader:
		r.CHROM = r.Chromosome
		r.START = r.Start_Position
		r.END   = r.End_Position
		r.NAME  = '%s:%s' % (r.CHROM, r.START)
		r.GENOTYPE = 'AA' if r.Tumor_Seq_Allele1 == r.Tumor_Seq_Allele2 \
						and r.Tumor_Seq_Allele1 == r.Reference_Allele else \
					 'BB' if r.Tumor_Seq_Allele1 == r.Tumor_Seq_Allele2 \
						and r.Tumor_Seq_Allele1 != r.Reference_Allele else 'AB'
		r.REFCOUNTS = r.t_ref_count
		r.VARCOUNTS = r.t_alt_count
		r.CASE      = r.Tumor_Sample_Barcode
		try:
			varcount  = float(r.VARCOUNTS)
			refcount  = float(r.REFCOUNTS)
			depth     = float(r.get('t_depth', 0))
			if depth == 0:
				depth = varcount + refcount
			r.VARFREQ = varcount / depth
		except (ValueError, TypeError, ZeroDivisionError):
			logger.warning('Variant %s drop due to unknown t_ref_count(%s) or t_alt_count(%s)' % (
				r.NAME,
				r.t_ref_count,
				r.t_alt_count
			))
			continue

		if r.CASE not in ret:
			ret[r.CASE] = path.join(outdir, r.CASE + '.muts.bed')
			writers[r.CASE] = TsvWriter(ret[r.CASE])
			writers[r.CASE].cnames = ['CHROM', 'START', 'END', 'NAME', 'GENOTYPE', 'REFCOUNTS', 'VARCOUNTS', 'CASE', 'VARFREQ']
		writers[r.CASE].write(r)

	for writer in writers.values():
		writer.close()
	reader.close()
	return ret

def PyCloneMutTSV2BED(pcfile):
	ret = {}
	writers = {}
	reader = TsvReader(pcfile)
	for r in reader:
		if ':' not in r.mutation_id:
			raise "`mutation_id` should end with `<chr>:<pos>`"
		chr, pos = r.mutation_id.split(':')[:-2]

		r.CHROM     = chr
		r.START     = pos
		r.END       = pos
		r.NAME      = '%s:%s' % (chr, pos)
		r.GENOTYPE  = r.genotype
		r.REFCOUNTS = r.ref_counts
		r.VARCOUNTS = r.var_counts
		r.CASE      = r.variant_case
		r.VARFREQ   = r.variant_freq

		if r.CASE not in ret:
			ret[r.CASE] = path.join(outdir, r.CASE + '.muts.bed')
			writers[r.CASE] = TsvWriter(ret[r.CASE])
			writers[r.CASE].cnames = ['CHROM', 'START', 'END', 'NAME', 'GENOTYPE', 'REFCOUNTS', 'VARCOUNTS', 'CASE', 'VARFREQ']
		writers[r.CASE].write(r)
	for writer in writers.values():
		writer.close()
	reader.close()
	return ret

def singleSampleCnVCF2BED(vcffiles):
	"""
	Convert single(paired)-sample VCFs to a BED file like:
	#CHROM  START   END	    GENOTYPE NORMAL_CN MINOR_CN MAJOR_CN CASE
	chr1    70820   70820   BB       2         0        2        NA12156
	"""

	ret = {}
	for vcffile in vcffiles:
		samples = shell.bcftools.query(l = vcffile).strip().splitlines()
		assert len(samples) <= 2
		samidx  = 0 if len(samples) == 1 or cnctrl in (1, -1, None) else 1
		sample  = samples[samidx]
		ret[sample] = path.join(outdir, sample + '.cn.bed')
		writer = TsvWriter(ret[sample])
		writer.cnames = ['CHROM', 'START', 'END', 'GENOTYPE', 'NORMAL_CN', 'MINOR_CN', 'MAJOR_CN', 'CASE']
		for line in shell.bcftools.query(
			_ = vcffile,
			s = sample,
			f = '%CHROM\t%POS\t%POS\t%GT\t2\t0\t2\t%SAMPLE\n'):

			items = line.split('\t')
			if items[3] in ('0/0', '0|0'):
				items[3] = 'AA'
			elif items[3] in ('0/1', '0|1'):
				items[3] = 'AB'
			elif items[3] in ('1/1', '1|1'):
				items[3] = 'BB'

			items[4], items[5], items[6] = defaultCNs(items[0], items[3])

			writer.write(items)
		writer.close()
	return ret

def multiSampleCnVCF2BED(vcffile):
	ret = {}
	writers = {}
	samples = shell.bcftools.query(l = vcffile).strip().splitlines()
	samples = [sample for i, sample in samples
		if cnctrl is None or i != (cnctrl if cnctrl >= 0 else len(samples) + cnctrl)]
	for sample in samples:
		ret[sample] = path.join(outdir, sample + '.cn.bed')
		writer = TsvWriter(sample)
		writer.cnames = ['CHROM', 'START', 'END', 'GENOTYPE', 'NORMAL_CN', 'MINOR_CN', 'MAJOR_CN', 'CASE']
		writers[sample] = writer
	for line in shell.bcftools.query(
		_ = vcffile,
		s = ','.join(samples),
		f = '[%CHROM\t%POS\t%POS\t%GT\t2\t0\t2\t%SAMPLE\n]'):

		items = line.split('\t')
		if items[3] in ('0/0', '0|0'):
			items[3] = 'AA'
		elif items[3] in ('0/1', '0|1'):
			items[3] = 'AB'
		elif items[3] in ('1/1', '1|1'):
			items[3] = 'BB'

		items[4], items[5], items[6] = defaultCNs(items[0], items[3])
		writer = writers[items[7]]
		writer.write(items)
	for writer in writers.values():
		writer.close()
	return ret

def PyCloneCnTSV2BED(pcfile):
	reader = TsvReader(pcfile)
	ret = {}
	writers = {}
	for r in reader:
		if ':' not in r.mutation_id:
			raise "`mutation_id` should end with `<chr>:<pos>`"
		chr, pos = r.mutation_id.split(':')[:-2]

		r.CHROM     = chr
		r.START     = pos
		r.END       = pos
		r.GENOTYPE  = r.genotype
		r.CASE      = r.variant_case
		r.NORMAL_CN = r.normal_cn
		r.MINOR_CN  = r.minor_cn
		r.MAJOR_CN  = r.major_cn
		if r.CASE not in ret:
			ret[r.CASE] = path.join(outdir, r.CASE + '.cn.bed')
			writers[r.CASE] = TsvWriter(ret[r.CASE])
			writers[r.CASE].cnames = ['CHROM', 'START', 'END', 'GENOTYPE', 'NORMAL_CN', 'MINOR_CN', 'MAJOR_CN', 'CASE']
		writers[r.CASE].write(r)
	for writer in writers.values():
		writer.close()
	reader.close()
	return ret

def GetPyCloneTsv(mutfile, outfile, cnfile = None):
	# mutbed:
	# #CHROM  START   END	    NAME      GENOTYPE REFCOUNTS VARCOUNTS CASE    VARFREQ
	# cnbed:
	# #CHROM  START   END	    GENOTYPE NORMAL_CN MINOR_CN MAJOR_CN CASE
	# outfile:
	# mutation_id	ref_counts	var_counts	normal_cn	minor_cn	major_cn	variant_case	variant_freq	genotype
	writer = TsvWriter(outfile)
	writer.cnames = ['mutation_id', 'ref_counts', 'var_counts', 'normal_cn', 'minor_cn', 'major_cn', 'variant_case', 'variant_freq', 'genotype']
	writer.writeHead()
	if not cnfile:
		reader = TsvReader(mutbed)
		reader.cnames = ['CHROM', 'START', 'END', 'NAME', 'GENOTYPE', 'REFCOUNTS', 'VARCOUNTS', 'CASE', 'VARFREQ']
		for r in reader:
			r.mutation_id = r.NAME
			r.ref_counts = r.REFCOUNTS
			r.var_counts = r.VARCOUNTS
			r.normal_cn, r.minor_cn, r.major_cn = defaultCNs(r.CHROM, r.GENOTYPE)
			r.variant_case = r.CASE
			r.variant_freq = r.VARFREQ
			r.genotype = r.GENOTYPE
			writer.write(r)
		writer.close()
		return

	for line in shell.bedtools.intersect(
		a = mutfile,
		b = cnfile,
		loj = True):

		# CHROM  START END NAME GENOTYPE REFCOUNTS VARCOUNTS CASE VARFREQ
		# 0      1     2   3    4        5         6         7    8
		# CHROM START END GENOTYPE NORMAL_CN MINOR_CN MAJOR_CN CASE
		# 9     10    11  12       13        14       15       16
		parts = line.split('\t')

		rec = TsvRecord()
		rec.mutation_id = parts[3]
		rec.ref_counts = parts[5]
		rec.var_counts = parts[6]
		rec.variant_case = parts[7]
		rec.variant_freq = parts[8]
		rec.genotype = parts[4]
		if parts[9] == '.':
			rec.normal_cn, rec.minor_cn, rec.major_cn = defaultCNs(parts[0], parts[4])
		else:
			rec.normal_cn, rec.minor_cn, rec.major_cn = parts[13:16]
		writer.write(rec)
	writer.close()

params  = Box(in_files = [])
mutbeds = {}
cnbeds  = {}
if path.isfile(inmuts):
	if inmuts.endswith('.vcf') or inmuts.endswith('.vcf.gz'):
		mutbeds = multiSampleMutVCF2BED(inmuts)
	elif inmuts.endswith('.maf') or inmuts.endswith('.maf.gz'):
		mutbeds = MAF2BED(inmuts)
	else:
		mutbeds = PyCloneMutTSV2BED(inmuts)
else:
	inmuts = inmuts.split(',')
	mutbeds = singleSampleMutVCF2BED(inmuts)

if incnvs and path.isfile(incnvs):
	if incnvs.endswith('.vcf') or incnvs.endswith('.vcf.gz'):
		cnbeds = multiSampleCnVCF2BED(incnvs)
	else:
		cnbeds = PyCloneCnTSV2BED(incnvs)
elif incnvs:
	incnvs = incnvs.split(',')
	cnbeds = singleSampleCnVCF2BED(incnvs)

for sample, mutbed in mutbeds.items():
	pcfile = path.join(outdir, sample + '.tsv')
	params.in_files.append(pcfile)
	GetPyCloneTsv(mutbed, pcfile, cnbeds.get(sample))

#PyClone run_analysis_pipeline --in_files A.tsv B.tsv C.tsv --working_dir pyclone_analysis
# create matplotlibrc file
with open(path.join(outdir, 'matplotlibrc'), 'w') as f:
	f.write('backend: Agg\n')
params.working_dir      = outdir
params.prior            = 'total_copy_number'
params.plot_file_format = 'svg'
params._env             = env
params._raise = False

c = shell.fg.pyclone.run_analysis_pipeline(**params)
if c.rc != 0:
	# Let it go if 'No mutations found in common across samples'
	exit(0)

# annotate tables/loci.tsv with genes
"""
mutation_id	sample_id	cluster_id	cellular_prevalence	cellular_prevalence_std	variant_allele_frequency
chr10:102776870	AK_nevus_1	0	0.656	0.127	0.417
chr10:102776870	AK_nevus_2	0	0.432	0.126	0.333
"""
locifile = path.join(outdir, 'tables', 'loci.tsv')
locifilebed = path.join(outdir, 'tables', 'loci.bed')
reader = TsvReader(locifile)
writer = TsvWriter(locifilebed)
writer.cnames = ['CHR', 'START', 'END'] + reader.cnames
for r in reader:
	r.CHR, r.START = r.mutation_id.split(':')
	r.END = r.START
	writer.write(r)
reader.close()
writer.close()

writer = TsvWriter(locifile)
writer.cnames = reader.cnames + ['gene']
writer.writeHead()
for line in shell.bedtools.intersect(a = locifilebed, b = refgene, loj = True, _iter = True):
	parts = line.split('\t')
	r = parts[3:9] + [parts[17].split('; ')[0][9:-1]]
	writer.write(r)
writer.close()





"""picard utilities"""
from pyppl import Proc, Box
from . import params
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pMarkDuplicates():
	"""
	@name:
		pMarkDuplicates
	@description:
		Identifies duplicate reads.
		This tool locates and tags duplicate reads in a BAM or SAM file, where duplicate reads are defined as originating from a single fragment of DNA. Duplicates can arise during sample preparation e.g. library construction using PCR. See also EstimateLibraryComplexity for additional notes on PCR duplication artifacts. Duplicate reads can also result from a single amplification cluster, incorrectly detected as multiple clusters by the optical sensor of the sequencing instrument. These duplication artifacts are referred to as optical duplicates.
		The MarkDuplicates tool works by comparing sequences in the 5 prime positions of both reads and read-pairs in a SAM/BAM file. An BARCODE_TAG option is available to facilitate duplicate marking using molecular barcodes. After duplicate reads are collected, the tool differentiates the primary and duplicate reads using an algorithm that ranks reads by the sums of their base-quality scores (default method).
		The tool's main output is a new SAM or BAM file, in which duplicates have been identified in the SAM flags field for each read. Duplicates are marked with the hexadecimal value of 0x0400, which corresponds to a decimal value of 1024. If you are not familiar with this type of annotation, please see the following [blog post](https://www.broadinstitute.org/gatk/blog?id=7019) for additional information.
		Although the bitwise flag annotation indicates whether a read was marked as a duplicate, it does not identify the type of duplicate. To do this, a new tag called the duplicate type (DT) tag was recently added as an optional output in the 'optional field' section of a SAM/BAM file. Invoking the TAGGING_POLICY option, you can instruct the program to mark all the duplicates (All), only the optical duplicates (OpticalOnly), or no duplicates (DontTag). The records within the output of a SAM/BAM file will have values for the 'DT' tag (depending on the invoked TAGGING_POLICY), as either library/PCR-generated duplicates (LB), or sequencing-platform artifact duplicates (SQ). This tool uses the READ_NAME_REGEX and the OPTICAL_DUPLICATE_PIXEL_DISTANCE options as the primary methods to identify and differentiate duplicate types. Set READ_NAME_REGEX to null to skip optical duplicate detection, e.g. for RNA-seq or other data where duplicate sets are extremely large and estimating library complexity is not an aim. Note that without optical duplicate counts, library size estimation will be inaccurate.
		MarkDuplicates also produces a metrics file indicating the numbers of duplicates for both single- and paired-end reads.
		The program can take either coordinate-sorted or query-sorted inputs, however the behavior is slightly different. When the input is coordinate-sorted, unmapped mates of mapped records and supplementary/secondary alignments are not marked as duplicates. However, when the input is query-sorted (actually query-grouped), then unmapped mates and secondary/supplementary reads are not excluded from the duplication test and can be marked as duplicate reads.
		If desired, duplicates can be removed using the REMOVE_DUPLICATE and REMOVE_SEQUENCING_DUPLICATES options.
	@input:
		`infile:file`:  The bam file
	@output:
		`outfile:file`: The marked bam file
	@args:
		`picard`:     The picard executable, default: "picard"
		`params`:  Other parameters for picard MarkDuplicates, default: ""
		`tmpdir`:  The tmpdir to use. Default: /tmp
	@requires:
		[picard](https://broadinstitute.github.io/picard/)
	"""
	pMarkDuplicates = Proc()
	pMarkDuplicates.input  = "infile:file"
	pMarkDuplicates.output = "outfile:file:{{infile | fn | lambda x: __import__('re').sub(r'(\\.sort|\\.sorted)?$', '', x)}}.dedup.bam"
	pMarkDuplicates.args   = { "picard": "picard", "tmpdir": "/tmp", "params": "", "javamem": "-Xms1g -Xmx8g" }
	pMarkDuplicates.script = """
	tmpdir="{{args.tmpdir}}/{{proc.id}}_{{#}}_{{infile | fn}}"
	mkdir -p "$tmpdir"
	mfile="{{job.outdir}}/{{infile | fn}}.metrics.txt"
	{{args.picard}} MarkDuplicates {{args.javamem}} -Djava.io.tmpdir="$tmpdir" TMP_DIR="$tmpdir" I="{{infile}}" O="{{outfile}}" M="$mfile" {{args.params}}
	rm -rf "$tmpdir"
	"""
	return pMarkDuplicates

@procfactory
def _pAddOrReplaceReadGroups():
	"""
	@name:
		pAddOrReplaceReadGroups
	@description:
		Replace read groups in a BAM file.This tool enables the user to replace all read groups in the INPUT file with a single new read group and assign all reads to this read group in the OUTPUT BAM file.
		For more information about read groups, see the [GATK Dictionary entry](https://www.broadinstitute.org/gatk/guide/article?id=6472).

		This tool accepts INPUT BAM and SAM files or URLs from the Global Alliance for Genomics and Health (GA4GH) (see http://ga4gh.org/#/documentation).
	@input:
		`infile:file`:  The bam file
		`rg`:           The read group information. For example:
			- "RGID=4 RGLB=lib1 RGPL=illumina RGPU=unit1 RGSM=20"
	@output:
		`outfile:file`: The bam file with read group added
	@args:
		`picard`:     The picard executable, default: "picard "
		`params`:  Other parameters for picard AddOrReplaceReadGroups, default: ""
	@requires:
		[picard](https://broadinstitute.github.io/picard/)
	"""
	pAddOrReplaceReadGroups = Proc()
	pAddOrReplaceReadGroups.input  = "infile:file, rg"
	pAddOrReplaceReadGroups.output = "outfile:file:{{infile | fn}}.rg.bam"
	pAddOrReplaceReadGroups.args   = { "picard": "picard", "params": "" }
	pAddOrReplaceReadGroups.script = """
	rg="{{rg}}"
	if [[ "$rg" != *"RGPL="* ]]; then rg="$rg RGPL=illumina"; fi
	if [[ "$rg" != *"RGPU="* ]]; then rg="$rg RGPU=unit1"; fi
	if [[ "$rg" != *"RGLB="* ]]; then rg="$rg RGLB=lib1"; fi
	{{args.picard}} AddOrReplaceReadGroups I="{{infile}}" O="{{outfile}}" $rg {{args.params}}
	"""
	return pAddOrReplaceReadGroups

@procfactory
def _pCreateSequenceDictionary():
	"""
	@name:
		pCreateSequenceDictionary
	@description:
		Creates a sequence dictionary for a reference sequence. This tool creates a sequence dictionary file (with ".dict" extension) from a reference sequence provided in FASTA format, which is required by many processing and analysis tools. The output file contains a header but no SAMRecords, and the header contains only sequence records.
		The reference sequence can be gzipped (both .fasta and .fasta.gz are supported).
	@input:
		`infile:file`:  The fasta file
	@output:
		`outfile:file`: The same fasta file, but with dict file created
	@args:
		`picard`:     The picard executable, default: "picard"
		`params`:  Other parameters for picard CreateSequenceDictionary, default: ""
	@requires:
		[picard](https://broadinstitute.github.io/picard/)
	"""
	pCreateSequenceDictionary = Proc()
	pCreateSequenceDictionary.input  = "infile:file"
	pCreateSequenceDictionary.output = "outfile:file:{{infile | bn}}"
	pCreateSequenceDictionary.args   = { "picard": "picard", "params": "" }
	pCreateSequenceDictionary.script = """
	link="{{job.outdir}}/{{infile | bn}}"
	ln -s "{{infile}}" "$link"
	{{args.picard}} CreateSequenceDictionary R="$link" {{args.params}}
	"""
	return pCreateSequenceDictionary

@procfactory
def _pCollectWgsMetrics():
	"""
	@name:
		pCollectWgsMetrics
	@description:
		Collect metrics about coverage and performance of whole genome sequencing (WGS) experiments.
		This tool collects metrics about the fractions of reads that pass base- and mapping-quality filters as well as coverage (read-depth) levels for WGS analyses. Both minimum base- and mapping-quality values as well as the maximum read depths (coverage cap) are user defined.

		Note: Metrics labeled as percentages are actually expressed as fractions!
	@input:
		`infile:file`:  The bam file
	@output:
		`outfile:file`: The metrics file
	@args:
		`picard`:     The picard executable, default: "picard"
		`params`:  Other parameters for `picard CollectWgsMetrics`, default: ""
		`reffile`: The reference file, default: ""
	@requires:
		[picard](https://broadinstitute.github.io/picard/)
	"""
	pCollectWgsMetrics = Proc()
	pCollectWgsMetrics.input  = "infile:file"
	pCollectWgsMetrics.output = "outfile:file:{{infile | bn}}.metrics.txt"
	pCollectWgsMetrics.args   = { "picard": "picard", "params": "", "reffile": "" }
	pCollectWgsMetrics.script = """
	if [[ -z "{{args.reffile}}" ]]; then
		echo "Reference file not specified!" 1>&2
		exit 1
	fi
	{{args.picard}} CollectWgsMetrics I="{{infile}}" O="{{outfile}}" R="{{args.reffile}}" {{args.params}}
	"""
	return pCollectWgsMetrics

@procfactory
def _pSortSam():
	"""
	@name:
		pSortSam
	@description:
		Use `picard SortSam` to sort sam or bam file
	@input:
		`infile:file`:  The sam or bam file to be sorted
	@output:
		`outfile:file`: The sorted sam or bam file
	@args:
		`picard`:     The picard executable, default: "picard"
		`order`:   The sort order, default: coordinate. Possible: unsorted, queryname, coordinate, duplicate
		`outtype`: The type of output file, sam or bam. Default: bam
		`params`:  Other parameters for `picard SortSam`, default: ""
		`tmpdir`:  The tmpdir to use. Default: /tmp
		`javamem`: The memory for java vm. Default: "-Xms1g -Xmx8g"
	@requires:
		[picard](http://broadinstitute.github.io/picard/command-line-overview.html)
	"""
	pSortSam = Proc()
	pSortSam.input  = "infile:file"
	pSortSam.output = "outfile:file:{{infile | fn}}.sorted.{{args.outtype}}"
	pSortSam.args   = { "picard": "picard", "order": "coordinate", "outtype": "bam", "params": "", "tmpdir": "/tmp", "javamem": "-Xms1g -Xmx8g" }
	pSortSam.script = """
	tmpdir="{{args.tmpdir}}/{{proc.id}}_{{proc.tag}}_{{#}}_{{infile | fn}}"
	mkdir -p "$tmpdir"
	{{args.picard}} SortSam {{args.javamem}} -Djava.io.tmpdir="$tmpdir" TMP_DIR="$tmpdir" {{args.params}} I="{{infile}}" O="{{outfile}}" SORT_ORDER={{args.order}}
	rm -rf "$tmpdir"
	"""
	return pSortSam

@procfactory
def _pIndexBam():
	"""
	@name:
		pIndexBam
	@description:
		Use `picard BuildBamIndex` to index bam file
	@input:
		`infile:file`:  The bam file
	@output:
		`outfile:file`: The same bam file (link) but with .bai file in `proc.outdir`
	@args:
		`picard`:    The picard executable, default: "picard"
		`params`:  Other parameters for `picard BuildBamIndex`, default: "-Xms1g -Xmx8g"
	@requires:
		[picard](http://broadinstitute.github.io/picard/command-line-overview.html)
	"""
	pIndexBam = Proc()
	pIndexBam.input  = "infile:file"
	pIndexBam.output = "outfile:file:{{infile | bn}}.bai"
	pIndexBam.args   = { "picard": "picard", "params": "-Xms1g -Xmx8g" }
	pIndexBam.script = """
	{{args.picard}} BuildBamIndex I="{{infile}}" O="{{outfile}}" {{args.params}}
	"""
	return pIndexBam

@procfactory
def _pCollectOxoGMetrics():
	"""
	@name:
		pCollectOxoGMetrics
	@description:
		Collect metrics to assess oxidative artifacts.This tool collects metrics quantifying the error rate resulting from oxidative artifacts. For a brief primer on oxidative artifacts, see the GATK Dictionary.
		This tool calculates the Phred-scaled probability that an alternate base call results from an oxidation artifact. This probability score is based on base context, sequencing read orientation, and the characteristic low allelic frequency. Please see the following reference for an in-depth discussion of the OxoG error rate.
		Lower probability values implicate artifacts resulting from 8-oxoguanine, while higher probability values suggest that an alternate base call is due to either some other type of artifact or is a real variant.
	@input:
		`infile:file`: The input bam file
	@output:
		`outfile:file`: The output file, Default: `{{i.infile | fn2}}.oxoG_metrics.txt`
	@args:
		`ref`   : The reference genome, Default: `<params.ref>`
		`picard`: The path to picard, Default: `<params.picard>`
		`params`: Other parameters for picard, Default: `Box()`
	"""
	pCollectOxoGMetrics             = Proc(desc = 'Collect metrics to assess oxidative artifacts.')
	pCollectOxoGMetrics.input       = 'infile:file'
	pCollectOxoGMetrics.output      = 'outfile:file:{{i.infile | fn2}}.oxoG_metrics.txt'
	pCollectOxoGMetrics.args.ref    = params.ref.value
	pCollectOxoGMetrics.args.picard = params.picard.value
	pCollectOxoGMetrics.args.params = Box()
	pCollectOxoGMetrics.lang        = params.python.value
	pCollectOxoGMetrics.script      = "file:scripts/picard/pCollectOxoGMetrics.py"
	return pCollectOxoGMetrics




"""Download/Get data from Websites instead of APIs"""
from pyppl import Proc
from . import params
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

# Web utils
@procfactory
def _pDownloadForm():
	"""
	@name:
		pDownloadForm
	@description:
		Download results by submitting a form, supporting pagination.
	@input:
		`url`   : the URL contains the form
		`data`  : the data used to fill the form (JSON string or transformed from dict by json.dumps).
		`submit`: the submit button to submit the form (use Xpath).
		`next`  : the button for next page (use Xpath)
	@output:
		`outdir:file`: The directory saves the results
	@args:
		`interval`: seconds to wait between fetching each page. Default: 1
	@requires:
		[`Splinter`](https://splinter.readthedocs.io/en/latest/index.html)
		[`Phantomjs`](http://phantomjs.org/)
	"""
	pDownloadForm               = Proc (desc="Download results by submitting to a form")
	pDownloadForm.input         = "url, data, submit, next"
	pDownloadForm.output        = "outdir:dir:{{i.url | bn | lambda x: x if x else 'outdir'}}"
	pDownloadForm.args.interval = 1
	pDownloadForm.lang          = params.python.value
	pDownloadForm.script        = "file:scripts/web/pDownloadForm.py"
	return pDownloadForm

@procfactory
def _pDownloadGet(alias = 'pDownload'):
	"""
	@name:
		pDownloadGet
	@description:
		Download results by urls.
	@input:
		`url`: the URLs to download
	@output:
		`outfile:file`: The output file
	"""
	pDownloadGet = Proc (desc="Download from URLs")
	pDownloadGet.input  = "url"
	pDownloadGet.output = "outfile:file:{{i.url | bn | lambda x: x if x else 'outfile' | .replace('?', '__Q__').replace('&', '__N__')}}"
	pDownloadGet.lang   = params.python.value
	pDownloadGet.script = "file:scripts/web/pDownloadGet.py"
	return pDownloadGet

@procfactory
def _pDownloadPost():
	"""
	@name:
		pDownloadPost
	@description:
		Download results by POST.
	@input:
		`url` : the URLs to download
		`data`: the POST data.
	@output:
		`outfile:file`: The output file
	"""
	pDownloadPost = Proc (desc="Download from URLs")
	pDownloadPost.input  = "url, data"
	pDownloadPost.output = "outfile:file:{{i.url | bn | lambda x: x if x else 'outfile'}}"
	pDownloadPost.lang   = params.python.value
	pDownloadPost.script = "file:scripts/web/pDownloadPost.py"
	return pDownloadPost




from os import path
from pyppl import Box
from bioprocs.utils import shell2 as shell
from bioprocs.utils.reference import bamIndex

infiles = {{i.infiles | repr}}
baits   = {{args.baits | repr}}
accfile = {{args.accfile | repr}}
ref     = {{args.ref | repr}}
params  = {{args.params | repr}}
prefix  = {{i.infiles | fs2name | quote}}
outdir  = {{job.outdir | quote}}
cnvkit  = {{args.cnvkit | quote}}
nthread = {{args.nthread | repr}}

for infile in infiles:
	bamIndex(infile)

shell.load_config(cnvkit = dict(
	_exe = cnvkit,
	_env = dict(
		OPENBLAS_NUM_THREADS = str(nthread),
		OMP_NUM_THREADS      = str(nthread),
		NUMEXPR_NUM_THREADS  = str(nthread),
		MKL_NUM_THREADS      = str(nthread)
	),
	_cwd = outdir
))

# generate target file
if baits[-4:] in ('.gff', '.gtf'):
	#https://github.com/pwwang/pygff
	from gff import Gff
	baitfile = path.join(outdir, path.basename(baits[:-4]) + '.bait.bed')
	gff = Gff(baits)
	with open(baitfile, 'w') as f:
		for g in gff:
			f.write("{chrom}\t{start}\t{end}\t{name}\n".format(
				chrom = g['seqid'],
				start = g['start'],
				end   = g['end'],
				name  = g['attributes']['gene_id'],
			))
	baits = baitfile

params_t   = params.target
params_t.o = path.join(outdir, prefix + '.bed')
shell.fg.cnvkit.target(baits, **params_t)

# generate access file
if not accfile:
	accfile    = path.join(outdir, prefix + '.access.bed')
	params_a   = params.access
	params_a.o = accfile
	shell.fg.cnvkit.access(ref, **params_a)

# autobin
params_b   = params.autobin
params_b.t = params_t.o
params_b.g = accfile
shell.fg.cnvkit.autobin(*infiles, **params_b)



from os import path, remove
from glob import glob
from bioprocs.utils import runcmd, cmdargs

indir   = {{i.indir | quote}}
outfile = {{o.outfile | quote}}
plink   = {{args.plink | quote}}
gz      = {{args.gz | repr}}
samid   = {{args.samid | quote}}
chroms  = {{args.chroms | repr}}

recode = ['vcf-fid' if samid == 'fid' else 'vcf-iid' if samid == 'iid' else 'vcf']
recode.append('tab')
if gz:
	outfile = outfile[:-3]

bedfile = glob(path.join(indir, '*.bed'))[0]
input   = path.splitext(bedfile)[0]
output  = path.splitext(outfile)[0]
outtmp  = output + '-tmp'

params = {
	'bfile' : input,
	'recode': recode,
	'out'   : outtmp
}
cmd = '%s %s 1>&2' % (plink, cmdargs(params, equal = ' '))
runcmd(cmd)

with open(outtmp + '.vcf', 'r') as fin, open(outfile, 'w') as fout:
	for line in fin:
		if line.startswith('#'):
			fout.write(line)
		else:
			items = line.split()
			if items[0] in chroms:
				items[0] = chroms[items[0]]
			elif 'chr' + items[0] in chroms:
				items[0] = chroms[items[0]][3:] if chroms[items[0]].startswith('chr') else chroms[items[0]]
			fout.write('\t'.join(items) + '\n')

remove(outtmp + '.vcf')
if gz:
	runcmd(['bgzip', outfile])




"""Some statistic processes"""
from pyppl import Proc, Box
from . import params, rimport
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pStats():
	"""
	@name:
		pStats
	"""
	return Box(
		desc   = 'Data statistics',
		lang   = params.Rscript.value,
		input  = 'infile:file',
		output = 'outdir:dir:{{i.infile | stem}}.stats',
		args   = Box(inopts  = Box(rnames = True, cnames = True, delimit = "\t"),
			types   = Box(),
			groups  = [],
			ignore  = [],
			devpars = Box(res = 300, height = 2000, width = 2000))
	)

@procfactory
def _pMetaPval():
	"""
	@name:
		pMetaPval
	@description:
		Calculate a meta-pvalue using different methods
	@input:
		`infile:file`: The infile containing multiple pvalues for each entries.
			- Could be two types (see `args.intype`)
			- `matrix`: A matrix with rows as entries (rownames are optional), columns as cases
				```
						Case1   Case2   ...  CaseN
				Entry1  1.33e-2 NA      ...  1.77e-10
				Entry2  2.66e-2 4.22e-5 ...  1.71e-3
				... ...
				EntryM  NA      0.00013 ...  4.11e-3
				```
			- `melt`: Rows are entries from each case
			`args.rnames` should be `False`, but entry names are required in the first column,
			column names are optional (have to be 3 columns)
				```
				Entry   Pvalue   Case
				Entry1  1.33e-2  Case1
				Entry1  1.77e-10 CaseN
				Entry2  2.66e-2  Case1
				Entry2  4.22e-5  Case2
				Entry2  1.71e-3  CaseN
				... ...
				EntryM  0.00013  Case2
				EntryM  4.11e-3  CaseN
				```
	@output:
		`outfile:file`: The output file containing the meta-pvalues. Default: `{{i.infile | fn}}.meta{{i.infile | ext}}`
	@args:
		`intype`: The type of the input file. Default: `matrix` (see `i.infile`)
		`inopts`: The input options to read the input file. Default: `Box(rnames = True, cnames = True)`
		`method`: The method used to calculate the meta-pvalue. Default: sumlog (Fisher's method, aka `fisher`)
			- Other available methods: logitp, sumz, votep, sump, meanp and wilkinsonp
			- See: https://www.rdocumentation.org/packages/metap/versions/0.8
		`na`    : How to deal with `NA` p-values. Default: `skip` (just don't count it)
			- Or a numeric value to replace it with (e.g.: `1`).
	@requires:
		[`r-matep`](https://www.rdocumentation.org/packages/metap/)
	"""
	pMetaPval              = Proc(desc = 'Calculate meta p-values.')
	pMetaPval.input        = 'infile:file'
	pMetaPval.output       = 'outfile:file:{{i.infile | fn}}.meta{{i.infile | ext}}'
	pMetaPval.args.intype  = 'matrix'
	pMetaPval.args.inopts  = Box(rnames = True, cnames = True)
	pMetaPval.args.method  = 'sumlog'
	pMetaPval.args.na      = 'skip' #0, 1
	pMetaPval.envs.rimport = rimport
	pMetaPval.lang         = params.Rscript.value
	pMetaPval.script       = "file:scripts/stats/pMetaPval.r"
	return pMetaPval

@procfactory
def _pSurvival():
	"""
	@input:
		infile: The input file (header is required).
			- col1: rownames if args.inopts.rnames = True
			- col2: the survival time
			- col3: the status. 0/1 for alive/dead or 1/2 for alive dead
			- col4: var1.
			- ... other variables
			- Note that covariales should be put in covfile, each variable in this file will be analyzed separately.
		covfile: The covariate file. If not provided, use `args.covfile`
	@output:
		outfile: The outfile containing the pvalues and other details.
		outdir : The output directory containing the pval files and plots, including:
			- `outfile`
			- cutplot: `<feature>.cut.png`, only available when `args.cutting = 'maxstat'`
			- forest plot: `<feature>.forest.png`, only available using cox
			- data with subpopulations: `<feature>.data.txt`
	@args:
		`inunit`    : The time unit in input file. Default: days
		`outunit`   : The output unit for plots. Default: days
		`inopts`    : The options for input file
			- `rnames`: Whether input file has row names.
		`combine`   : Whether combine variables in the same plot.
			- `nrow`: The number of rows. Default: 1
			- `ncol`: The number of cols. Default: 1
		`devpars`   : The device parameters for png.
			- The height and width are for each survival plot.
			- If args.combine is True, the width and height will be multiplied by `max(combine.ncol, combine.nrow)`
		`covfile`   : The covariant file. Require rownames in both this file and input file.
		`ggs`       : Extra ggplot2 elements for main plot. `ggs$table` is for the risk table.
		`params`    : The params for `ggsurvplot`. Extra arguments:
			- cut: Method to cut the continous value into subpopulations. One of mean, median, q25, q75, asis and maxstat (default).
			- cutlabels: Labels for the subpopulations (from small to large).
		`method`    : The method to do survival analysis. km (Kaplan Merier) or cox.
	@requires:
		[`r-survival`](https://rdrr.io/cran/survival/)
		[`r-survminer`](https://rdrr.io/cran/survminer/)
	"""
	return Box(desc = "Survival analysis",
		lang   = params.Rscript.value,
		input  = 'infile:file, covfile:file',
		output = [
			'outfile:file:{{i.infile | fn2}}.dir/{{i.infile | fn2}}.survival.txt',
			'outdir:dir:{{i.infile | fn2}}.dir'
		],
		args = Box(inunit = 'days', # months, weeks, years
			outunit   = 'days',
			method    = 'cox', # or km
			covfile   = None,
			inopts    = Box(rnames = True),
			# params for arrange_ggsurvplots.
			# Typically nrow or ncol is set.
			# If combine.ncol = 3, that means {ncol: 3, nrow: 1}.
			# If ncol is not set, then it defaults to 1.
			# If empty, the figures will not be combined
			combine   = Box(ncol = 2),
			devpars   = Box(res = 300, height = 2000, width = 2000),
			params    = Box(),
			ggs       = Box(table = Box())))

@procfactory
def _pPostSurvival():
	"""
	@name:
		pPostSurvival
	@description:
		Statistic comparison between groups after survival analysis.
	@input:
		`infile:file`: The result file from `pSurvival`
		`survfile:file`: The survival data. See format of infile of `pSurvival`
	@output:
		`outfile:file`: The output excel file.
	@args:
		`covfile`: The covariant file. Require rownames in both this file and input file.
		`methods`: A list of testing methods
			- `wilcox`: Wilcox rank sum test
			- `t`: t-test
			- `chisq`: chisquare-test
		`inopts`: The input options for `i.survfile`.
			- `rnames`: whether the file has row names. This has to be True if `args.covfile` provided.
	"""
	pPostSurvival              = Proc(desc = "Post survival analysis: statistics on variables in different groups")
	pPostSurvival.input        = 'infile:file, survfile:file'
	pPostSurvival.output       = 'outfile:file:{{i.infile | fn2}}.stats.xlsx'
	pPostSurvival.args.chi2n   = 10
	pPostSurvival.args.inopts  = Box(rnames = True)
	pPostSurvival.args.covfile = None
	pPostSurvival.envs.rimport = rimport
	pPostSurvival.lang         = params.Rscript.value
	pPostSurvival.script       = "file:scripts/stats/pPostSurvival.r"
	return pPostSurvival

@procfactory
def _pBin():
	"""
	@name:
		pBin
	@description:
		Bin the data in columns.
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The output file. Default: `{{i.infile | stem}}.binned{{i.infile | ext}}`
	@args:
		`inopts`: The input options.
			- `delimit`: The delimiter. Default: `\t`
			- `rnames`: Whether input file has row names. Default: `False`
			- `cnames`: Whether input file has column names. Default: `True`
			- Other arguments available for `read.table`
		`binopts`: The default bin options:
			- `nbin`: Number of bins.
			- `step`: The step of binning.
			- `nan`:  What to do if the value is not a number. Default: `skip`
				- `skip/keep`: Keep it
				- `as0`: Treat it as 0
			- `out`: The out value. Default: `step`
				- `step`: Use the step breaks
				- `lower/min`: Use the min value of the records in the bin
				- `upper/max`: Use the max value of the records in the bin
				- `mean`: Use the mean value of the records in the bin
				- `median`: Use the median value of the records in the bin
				- `binno`: Use the bin number (empty bins will be skipped).
		`cols`: The detailed bin options for each column.
			- If not provided (`None`), all columns will use `binopts`.
			- If column specified, only the specified column will be binned.
			- Column indices can be used. It's 1-based.
	"""
	pBin              = Proc(desc = "Bin the data")
	pBin.input        = 'infile:file'
	pBin.output       = 'outfile:file:{{i.infile | stem}}.binned{{i.infile | ext}}'
	pBin.args.inopts  = Box(delimit = "\t", rnames = False, cnames = True)
	pBin.args.binopts = Box(nbin = None, step = None, nan = 'skip', out = 'step') # lower/min, upper/max, mean, median, binno
	pBin.args.cols    = None
	pBin.envs.rimport = rimport
	pBin.lang         = params.Rscript.value
	pBin.script       = "file:scripts/stats/pBin.r"
	return pBin

@procfactory
def _pQuantileNorm():
	"""
	@name:
		pQuantileNorm
	@description:
		Do quantile normalization
	@input:
		`infile:file`: The input matrix
	@output:
		`outfile:file`: The output matrix. Default: `{{i.infile | bn}}`
	"""
	pQuantileNorm              = Proc(desc = 'Do quantile normalization')
	pQuantileNorm.input        = 'infile:file'
	pQuantileNorm.output       = 'outfile:file:{{i.infile | bn}}'
	pQuantileNorm.args.inopts  = Box(rnames = True, cnames = True, delimit = "\t", skip = 0)
	pQuantileNorm.envs.rimport = rimport
	pQuantileNorm.lang         = params.Rscript.value
	pQuantileNorm.script       = "file:scripts/stats/pQuantileNorm.r"
	return pQuantileNorm

@procfactory
def _pChiSquare():
	"""
	@name:
		pChiSquare
	@description:
		Do chi-square test.
	@input:
		`infile:file`: The input file.
	@output:
		`outfile:file` : The output file containing Xsquare, df, pval and method
		`obsvfile:file`: The observation matrix
		`exptfile:file`: The expectation matrix
	@args:
		`intype`: The type of the input file:
			- `count` (default): The contingency table
			```
			#         | Disease | Healthy |
			# --------+---------+---------+
			#   mut   |   40    |   12    |
			# non-mut |   23    |   98    |
			# --------+---------+---------+
			```
			- `raw`: The raw values:
			```
			# Contingency table rows: Mut, Non
			# Contingency table cols: Disease, Healthy
			#
			#         | S1 | S2 | ... | Sn |
			# --------+----+----+-----+----+
			# Disease | 1  | 0  | ... | 1  |
			# Healthy | 0  | 1  | ... | 0  |
			# --------+----+----+-----+----+
			# Mut     | 1  | 0  | ... | 1  |
			# Non     | 0  | 1  | ... | 0  |
			```
		`ctcols`: The colnames of contingency table if input file is raw values
			- You may also specify them in the head of the input file
	"""
	pChiSquare = Proc(desc = "Do chi-square test.")
	pChiSquare.input = "infile:file"
	pChiSquare.output = "outfile:file:{{i.infile | fn2}}.chi2.txt, obsvfile:file:{{i.infile | fn2}}.obsv.txt, exptfile:file:{{i.infile | fn2}}.expt.txt"
	pChiSquare.args.intype = 'cont' # raw
	pChiSquare.args.ctcols = ''
	pChiSquare.lang = params.Rscript.value
	pChiSquare.script = "file:scripts/stats/pChiSquare.r"
	return pChiSquare

@procfactory
def _pFisherExact():
	"""
	@name:
		pFisherExact
	@description:
		Do fisher exact test.
	@input:
		`infile:file`: The input file.
	@output:
		`outfile:file` : The output file containing confInt1, confInt2, oddsRatio, pval, alternative and method.
	@args:
		`intype`: The type of the input file:
			- `count` (default): The contingency table
			```
			#         | Disease | Healthy |
			# --------+---------+---------+
			#   mut   |   40    |   12    |
			# non-mut |   23    |   98    |
			# --------+---------+---------+
			```
			- `raw`: The raw values:
			```
			# Contingency table rows: Disease, Healthy
			# Contingency table cols: Mut, Non
			#
			#    | Disease Healthy | Mut  Non  |
			# ---+--------+--------+-----+-----+
			# S1 |    1   |    0   |  0  |  1  |
			# S2 |    0   |    1   |  1  |  0  |
			# .. |   ...  |   ...  | ... | ... |
			# Sn |    0   |    1   |  0  |  1  |
			#
			```
		`ctcols`: The colnames of contingency table if input file is raw values
			- You may also specify them in the head of the input file
	"""
	pFisherExact             = Proc(desc = "Do fisher exact test.")
	pFisherExact.input       = "infile:file"
	pFisherExact.output      = "outfile:file:{{i.infile | fn2}}.fexact.txt"
	pFisherExact.args.intype = 'cont' # raw
	pFisherExact.args.ctcols = []
	pFisherExact.lang        = params.Rscript.value
	pFisherExact.script      = "file:scripts/stats/pFisherExact.r"
	return pFisherExact

@procfactory
def _pPWFisherExact():
	"""
	@name:
		pPWFisherExact
	@description:
		Do pair-wise fisher exact test.
		Commonly used for co-occurrence/mutual-exclusivity analysis.
		P-value indicates if the pairs are significantly co-occurred or mutually exclusive.
		Co-occurrence: Odds ratio > 1
		Mutual-exclusivity: Odds ratio < 1
	@input:
		`infile:file`: The input file.
	@output:
		`outfile:file` : The output file containing confInt1, confInt2, oddsRatio, pval, qval, alternative and method.
	@args:
		`intype`: The type of the input file:
			- `pairs`: The contingency table
			```
			#
			# A+	B+	4
			# A-	B-	175
			# A+	B-	12
			# A-	B+	1
			#
			```
			- `raw` (default): The raw values:
			```
			#
			#    | A | B | ... | X |
			# ---+---+---+-----+---+
			# S1 | 1 | 0 | ... | 1 |
			# S2 | 0 | 1 | ... | 0 |
			# .. | 0 | 0 | ... | 1 |
			# Sn | 0 | 1 | ... | 1 |
			#
			```
		`padj`: The p-value adjustment method, see `p.adjust.methods` in R. Default: `BH`
		`rnames`: If the input file has rownames for `raw` input type.
	"""
	pPWFisherExact              = Proc(desc = "Do pair-wise fisher exact test.")
	pPWFisherExact.input        = "infile:file"
	pPWFisherExact.output       = "outfile:file:{{i.infile | fn2}}.pwfexact.txt"
	pPWFisherExact.args.rnames  = True
	pPWFisherExact.args.intype  = 'raw' # pairs
	pPWFisherExact.args.padj    = 'BH'
	pPWFisherExact.envs.rimport = rimport
	pPWFisherExact.lang         = params.Rscript.value
	pPWFisherExact.script       = "file:scripts/stats/pPWFisherExact.r"
	return pPWFisherExact

@procfactory
def _pMediation():
	"""
	@name:
		pMediation
	@description:
		Do mediation analysis
	@input:
		`infile:file`: The input file (a matrix or data.frame). Example:
			```
				V1   V2   V3
			S1   1    2    3
			S2   4    1    8
			... ...
			Sn   3    3    1
			```
		`casefile:file`: The mediation options. Example:
			```
			Case1   lm      V3~V2|V1
			Case2   lm,glm  V3~V1|V2
			```
			- No column names, but implies `Case`, 'Model' and `Formua`.
			- `\t` as delimiter.
			- This file is optional. If it is not provided, `args.case` is required.
			- If this file is provided, `args.case` is ignored
			- For different models, model for Mediator comes last. For Case2, the models will be:
				- lm(V3 ~ V2 + V1) and glm(V2 ~ V1)
	@output:
		`outfile:file`: The result file.
		`outdir:dir`  : The output directory containing output file and plots.
	@args:
		`inopts`: The options for input file. Default: `Box(cnames = True, rnames = True)`
			- `cnames`: Whether the input file has column names
			- `rnames`: Whether the input file has row names
		`medopts`: The options for mediation analysis.
			- `boot`: Use bootstrap?
			- `sims`: How many time simulations?
		`cov`: The covariate file. Default: ``
		`pval`: The pvalue cutoff. Default: `0.05`
		`fdr` : Method to calculate fdr. Use `False` to disable. Default: `True` (`BH`)
		`plot`: Parameters for `plot.mediate`? Use `False` to disable plotting. Default: `Box()`
			- Only case with pvalue < `args.pval` will be plotted.
			- To plot all cases, use `args.pval = 1`
		`nthread`: Number of threads to use for different cases. Default: `1`
		`devpars`: device parameters for the plot. Default: `Box(res = 300, width = 2000, height = 2000)`
		`case`   : Define cases, each case should have `model` and `fmula`.
			- If you only have one case, then it could be: `Box(model = 'lm', fmula = 'Y~X|M')`
			In this case, `{{i.infile | fn2}}` will be used as case name
			- For multiple cases, this should be a dict of cases:
			`Box(Case1 = Box(model='lm', fmula='Y~X|M'), Case2 = ...)`
	"""
	pMediation = Proc(desc = "Do mediation analysis.")
	pMediation.input  = 'infile:file, casefile:file'
	pMediation.output = [
		'outfile:file:{{i.infile | fn2}}.mediation/{{i.infile | fn2}}.mediation.txt',
		'outdir:dir:{{i.infile | fn2}}.mediation'
	]
	pMediation.args.inopts  = Box(cnames = True, rnames = True)
	pMediation.args.medopts = Box(boot = True, sims = 500)
	pMediation.args.cov     = ''
	pMediation.args.pval    = 0.05
	pMediation.args.fdr     = True # BH, or other methods for p.adjust
	pMediation.args.plot    = Box()
	pMediation.args.case    = Box(model = 'lm', fmula = 'Y~X|M')
	pMediation.args.nthread = 1
	pMediation.args.devpars = Box(res = 300, width = 2000, height = 2000)
	pMediation.envs.rimport = rimport
	pMediation.lang         = params.Rscript.value
	pMediation.script       = "file:scripts/stats/pMediation.r"
	return pMediation

@procfactory
def _pLiquidAssoc():
	"""
	@name:
		pLiquidAssoc
	@description:
		Do liquid association analysis
	@input:
		`infile:file`: The input file with input data, where LA will be done on rows.
			```
				S1   S2 ... ... Sn
			G1   1    2  ... ... 9
			G2   3    1  ... ... 1
			... ...
			Gm   9    2  ... ... 3
			```
		`casefile:file`: Defining the groups (X, Y, Z) and the cases. If case (3rd col) is not provided, all will be treated as one case.
			- Group "Z" is required. You can also specify group "X", then the rest will be group "Y"
			```
			G1   X   Case1
			G2   X   Case1
			Gx   Z   Case1
			Gy   Z   Case1
			```
	@output:
		`outfile:file`: The results of the analysis
		`outdir:dir`  : The output directory containing the result file and plots.
	@args:
		`inopts` : The options for reading input file
		`zcat`   : Whether the group "Z" is categorical. Default: `False`
			- If it is, then `stein lemma` is not suitable, we will calculate LA manually (`E(g'(z)`)
		`pval`   : The pval cutoff. Default: `0.05`
		`fdr`    : The method to calculate FDR. Use `False` to disable. Default: `True` (BH)
		`nthread`: The number of threads to use. Default: `1` (WCGNA requires)
		`plot`   : Whether do plotting or not. Default: `False`
		`devpars`: device parameters for the plot. Default: `Box(res = 300, width = 2000, height = 2000)`
	@requires:
		[r-fastLiquidAssociation](https://github.com/pwwang/fastLiquidAssociation)
	"""
	pLiquidAssoc        = Proc()
	pLiquidAssoc.input  = 'infile:file, casefile:file'
	pLiquidAssoc.output = [
		'outfile:file:{{i.infile | fn2}}.la/{{i.infile | fn2}}.la.txt',
		'outdir:dir:{{i.infile | fn2}}.la'
	]
	pLiquidAssoc.args.inopts = Box(cnames = True, rnames = True)
	pLiquidAssoc.args.zcat    = False
	pLiquidAssoc.args.pval    = 0.05
	pLiquidAssoc.args.fdr     = True # BH, or other methods for p.adjust
	pLiquidAssoc.args.fdrfor  = 'case' # all
	pLiquidAssoc.args.nthread = 1
	pLiquidAssoc.args.plot    = False
	pLiquidAssoc.args.ggs     = Box()
	pLiquidAssoc.args.devpars = Box(res = 300, width = 2000, height = 2000)
	pLiquidAssoc.envs.rimport = rimport
	pLiquidAssoc.lang         = params.Rscript.value
	pLiquidAssoc.script       = "file:scripts/stats/pLiquidAssoc.r"
	return pLiquidAssoc

@procfactory
def _pHypergeom():
	"""
	@name:
		pHypergeom
	@description:
		Do hypergeometric test.
	@input:
		`infile:file`: The input file, could be raw data (presence (1) and absence (0) of elements) or number of overlapped elements and elements in each category.
			- Set `args.intype` as `raw` if it is raw data. The population size `args.N` is required
			- Set `args.intype` as `numbers` (or any string except `raw`) if it is numbers. You can specified explicit header: `k` = overlapped elements, `m` = size of set 1, `n` = size of set 2 and `N` = the population size. If `N` not included, then `args.N` is required
	@output:
		`outfile:file`: The output file
	@args:
		`intype`: the type of input file. Default: `raw`. See `infile:file`
		`inopts`: The options for input file.
			- `cnames`: Whether the input file has column names
			- `rnames`: Whether the input file has row names
		`N`: The population size. Default: `None`
	"""
	pHypergeom              = Proc(desc = "Do hypergeometric test.")
	pHypergeom.input        = 'infile:file'
	pHypergeom.output       = 'outfile:file:{{i.infile | fn2}}.hypergeom.txt'
	pHypergeom.args.intype  = 'raw' # numbers
	pHypergeom.args.inopts  = Box(cnames = True, rnames = True)
	pHypergeom.args.N       = None
	pHypergeom.envs.rimport = rimport
	pHypergeom.lang         = params.Rscript.value
	pHypergeom.script       = "file:scripts/stats/pHypergeom.r"
	return pHypergeom

@procfactory
def _pChow():
	"""
	@name:
		pChow
	@description:
		Do Chow-Test
	@input:
		`infile:file`: The input file for data to do the regressions. Example:
			```
				X1  X2  X3  X4 ... Y
			S1  1   2   1   4  ... 9
			S2  2   3   1   1  ... 3
			... ...
			Sm  3   9   1   7  ... 8
			```
		`groupfile:file`: Specify the groups to compare. You may also specify the cases. The Chow-Test will be done between the group for each case. Example:
			```
					Case1	Case2
			S1	Group1	Group1
			S2	Group1	NA          # Exclude S2 in Case2
			S3	Group2	Group1
			... ...
			Sm	Group2	Group2
			```
			- In such case, the test will be done between Group1 and Group2 for Case1 and Case2, respectively.
			- Instances can be resued (Gt in the example)
			- If cases not provided, all will be treated as one case.
		`casefile:file`: Define the formula (which columns to use for each case). Example:
			```
			Case1	Y ~ X1
			Case2	Y ~ X2
			```
			- This file is optional, then formula `Y ~ .` will be used for all cases.
	@output:
		`outfile:file`: The result file of chow test. Default: `{{i.infile | fn}}.chow/{{i.infile | fn}}.chow.txt`
		`outdir:dir`: The output directory, containing the output file, results of regressions and plots.
	@args:
		`inopts`: The options for input file.
			- `cnames`: Whether the input file has column names. Default: `True`
			- `rnames`: Whether the input file has row names. Default: `True`
		`cov`: The covariate file. `inopts.rnames` required and this file should have row names too. Default: `''`
		`fdr`   : Calculate FDR or not. Use `False` to disable. If `True` will use `BH` method, otherwise, specify the method (see `R`'s `p.adjust`).
		`pval`: The pvalue cutoff. Default: `0.05`
		`plot`: Whether plot the regressions. Default: `False`
		`ggs` : The extra ggs for the plot.
		`devpars`: device parameters for the plot. Default: `Box(res = 300, width = 2000, height = 2000)`
	"""
	pChow              = Proc(desc = "Do Chow-Test")
	pChow.input        = 'infile:file, groupfile:file, casefile:file'
	pChow.output       = 'outfile:file:{{i.infile | fn}}.chow/{{i.infile | fn}}.chow.txt, outdir:dir:{{i.infile | fn}}.chow'
	pChow.args.inopts  = Box(cnames = True, rnames = True)
	pChow.args.cov     = '' # co-variates, inopts.rnames required, and must in same order
	pChow.args.pval    = 0.05
	pChow.args.fdr     = True
	pChow.args.plot    = True
	pChow.args.devpars = Box(res = 300, width = 2000, height = 2000)
	pChow.args.ggs     = Box()
	pChow.envs.rimport = rimport
	pChow.lang         = params.Rscript.value
	pChow.script       = "file:scripts/stats/pChow.r"
	return pChow

@procfactory
def _pAnovaModel():
	"""
	@name:
		pAnovaModel
	"""
	pAnovaModel              = Proc(desc = "Run anova test on model to test the significance with/without terms")
	pAnovaModel.input        = 'infile:file, casefile:file'
	pAnovaModel.output       = 'outfile:file:{{i.infile | fn}}.anova/{{i.infile | fn}}.anova.txt, outdir:dir:{{i.infile | fn}}.anova'
	pAnovaModel.args.model   = 'lm'
	pAnovaModel.args.fmula   = None
	pAnovaModel.args.inopts  = Box(cnames = True, rnames = True)
	pAnovaModel.args.cov     = '' # requires inopts.rnames
	pAnovaModel.envs.rimport = rimport
	pAnovaModel.lang         = params.Rscript.value
	pAnovaModel.script       = "file:scripts/stats/pAnovaModel.r"
	return pAnovaModel

@procfactory
def _pCorr():
	"""
	@input:
		infile: The input file of data to calculate correlations.
			- The columns are instances and rows variables when `args.byrow = True`.
			- Transposed otherwise.
		groupfile: The group file defining the groups of variables, then the correlation will be only calculated between variables in different groups.
			- First column is the variable names, and second is the groups. Only two groups allowed.
			- Overlaps allowed in different groups.
			- If this is not provided, `args.groupfile` will be checked.
			- If none of them provided, correlation will be calculated for each pair of variables.
			- We can also specify variables directly here by 'var1,var2;var3,var4'
	@output:
		outfile: The output file containing the correlation coefficients
		outdir : The output directory containing the outfile, pvalue file and the plot
	@args:
		outfmt: The output format. Could be `matrix` or `pairs`.
		method: The method used to calculate the correlation coefficient.
			- Could be `pearson`, `spearman` or `kendall`
		byrow: Whether the rows are variables or not.
		pval:  Whether output the pvalues as well.
			- Will be generated at `<outdir>/<i.infile | stem>.<args.method>.pval.txt`
		groupfile: The group file. See `in.groupfile`
		inopts: The input options:
			- `cnames`: Required if `args.byrow = False`. Whether the input file has header.
			- `rnames`: Required if `args.byrow = True`. Whether the input file has row names.
			- `delimit`: The separator of columns.
		plot   : Whether output a correlation plot.
		params : The params for `plot.heatmap` in `utils/plot.r`
		ggs    : The extra ggplot2 statements.
		devpars: The parameters for the plot device.
	@requires:
		R packages: `ggplot2` and `reshape`
	"""
	return Box(
		desc   = 'Correlation Coefficients between variables',
		lang   = params.Rscript.value,
		input  = 'infile:file, groupfile:var',
		output = [
			'outfile:file:{{i.infile | stem}}.{{args.method}}/{{i.infile | stem}}.{{args.method}}.txt',
			'outdir:dir:{{i.infile | stem}}.{{args.method}}'
		],
		args = Box(
			outfmt    = 'pairs',
			method    = 'pearson',
			byrow     = True,
			pval      = False,
			groupfile = None,
			inopts    = Box(cnames = True,	rnames = True,	delimit = "\t"),
			plot      = False,
			params    = Box(), # the parameters for plot.heatmap2
			devpars   = Box(height = 2000, width = 2000, res = 300)
		)
	)

@procfactory
def _pCorr2():
	"""
	@name:
		pCorr2
	@description:
		Calculate correlation coefficient between instances of two files
		Don't do it between instances within the same file.
	@input:
		`infile1:file`: The first file. See input of `pCorr`
		`infile2:file`: The second file.
			- must have same number of columns with `infile1`
	@output:
		`outfile:file`: The output file.
		`outdir:dir`  : The output directory containing output file and other files:
			- pvalues/fdr file and plots
	@args:
		`pval`  : Whether output pvalue. Default: `False`
		`fdr`   : Whether output qvalue. Default: `False`
		`outfmt`: The output format. `pairs` (default) or `matrix`
		`plot`  : Whether plot a heatmap or not. Default: `False`
		`params`: The params for `plot.heatmap` in `utils/plot.r`
		`ggs`:    The extra ggplot2 statements.
		`devpars`:The parameters for the plot device. Default: `Box(height = 2000, width = 2000, res = 300)`
	"""
	pCorr2        = Proc(desc = 'Calculate correlation coefficient between instances of two files')
	pCorr2.input  = 'infile1:file, infile2:file'
	pCorr2.output = [
		'outfile:file:{{i.infile1 | fn2}}-{{i.infile2 | fn2}}.corr/{{i.infile1 | fn2}}-{{i.infile2 | fn2}}.corr.txt',
		'outdir:dir:{{i.infile1 | fn2}}-{{i.infile2 | fn2}}.corr'
	]
	pCorr2.args.inopts1 = Box()
	pCorr2.args.inopts2 = Box()
	pCorr2.args.method  = 'pearson' # spearman, kendall
	pCorr2.args.pval    = False
	pCorr2.args.fdr     = False # method: 'BH'
	pCorr2.args.outfmt  = 'pairs' # matrix
	pCorr2.args.plot    = False
	pCorr2.args.params  = Box() # the parameters for plot.heatmap
	pCorr2.args.ggs     = Box() # extra ggplot statements
	pCorr2.args.devpars = Box(height = 2000, width = 2000, res = 300)
	pCorr2.envs.rimport = rimport
	pCorr2.lang         = params.Rscript.value
	pCorr2.script       = "file:scripts/stats/pCorr2.r"
	return pCorr2

@procfactory
def _pDiffCorr():
	"""
	@name:
		pDiffCorr
	@description:
		Test correlation differences using Fisher Z method.
	@input:
		`infile:file`: The entire dataset used to calculate correlations. Rownames and colnames are required. Example:
			```
					S1  S2  S3  S4 ... Sn
				G1  1   2   1   4  ... 9
				G2  2   3   1   1  ... 3
				... ...
				Gm  3   9   1   7  ... 8
			```
		`samfile:file`: The sample groups, between which you want to compare the correlations. You can also specify one sample to multiple groups, and assign with different cases. Example:
			```
				S1	Healthy
				S2	Healthy
				S3	Disease
				... ...
				Sn	Disease
			```
		`casefile:file`: Assign the cases to compare. If not provided, it will do for every possible combination. Example:
			```
				Healthy	Disease
			```
		`groupfile:file`: Specify groups for rows, then the correlation will be only done within the pairs, each of which is from different groups (only 2 allowed). If not provided, it will investigate for every possible row pairs. Example:
			```
				G1	Kinase
				G2	Kinase
				... ...
				Gm	TF
			```
	@output:
		`outfile:file`: The pairs under different cases that their correlations have been changed significantly
		`outdir:dir`: The output directory containing plots and other output files.
	@args:
		`inopts`: The input options for `infile`. `cnames` and `rnames` have to be `True`
		`method`: The method to calculate the correlation. Default: `pearson`
		`pval`  : The pvalue cutoff to define correlation change significance. Default: `0.05`
		`fdr`   : Calculate FDR or not. Use `False` to disable. If `True` will use `BH` method, otherwise, specify the method (see `R`'s `p.adjust`).
		`fdrfor`: Do FDR calculation for each case (`case`) or all instances (`all`). Default: `case`
		`plot`  : Plot the correlation for cases? Default: `False`
		`ggs`   : `ggs` items for the plot.
		`devpars`: The device parameters for the plot.
	"""
	pDiffCorr = Proc(desc = 'Test correlation differences.')
	pDiffCorr.input  = 'infile:file, samfile:file, casefile:file, groupfile:file'
	pDiffCorr.output = [
		'outfile:file:{{i.infile | fn2}}.diffcorr/{{i.infile | fn2}}.diffcorr.txt',
		'outdir:dir:{{i.infile | fn2}}.diffcorr'
	]
	pDiffCorr.args.inopts  = Box(cnames = True, rnames = True)
	pDiffCorr.args.method  = 'pearson' # spearman
	pDiffCorr.args.pval    = 0.05
	pDiffCorr.args.fdr     = True # BH
	pDiffCorr.args.fdrfor  = 'case'
	pDiffCorr.args.plot    = False
	pDiffCorr.args.ggs     = Box() # extra ggplot statements
	pDiffCorr.args.devpars = Box(height = 2000, width = 2000, res = 300)
	pDiffCorr.envs.rimport = rimport
	pDiffCorr.lang         = params.Rscript.value
	pDiffCorr.script       = "file:scripts/stats/pDiffCorr.r"
	return pDiffCorr

@procfactory
def _pBootstrap():
	"""
	@name:
		pBootstrap
	@description:
		Do bootstrapping resampling
	@input:
		`infile:file`: The input data file
	@output:
		`outfile:file`: The output file with the bootstrapped statistics values
			- depends on the `args.stats` function
		`outdir:dir`: The directory to save the outfile and figures.
	@args:
		`inopts`:  The options to read the input file. Default: `Box(cnames = True, rnames = True)`
		`params`:  Other parameters for `boot` function from R `boot` package
		`nthread`: # of threads(cores) to use. Default: `1`
		`n`: Sampling how many times? Default: `1000`
		`stats`: The function to generate statistics for output. Default: `function(x) x`
			- Default to use all data
			- This function can return a multiple statistics in a vector
			- The argument `x` is the data generate for each sampling.
			- Unlink the `statistic` argument from `boot`, to make it convenient, we don't put the `index` here.
		`plot`: Plot the statistics? Default: `all` (plot all statistics)
			- You may also specify indices. For example: `[1, 2]` to plot the 1st and 2nd statistics
			- Use `False` or `None` to disable plotting
		`devpars`: The device parameters for the plot.
	"""
	pBootstrap        = Proc(desc = 'Do bootstrapping')
	pBootstrap.input  = 'infile:file'
	pBootstrap.output = [
		'outfile:file:{{i.infile | fn2}}.boot/{{i.infile | fn2}}.boot.txt',
		'outdir:dir:{{i.infile | fn2}}.boot'
	]
	pBootstrap.args.inopts  = Box(cnames = True, rnames = True)
	pBootstrap.args.params  = Box()
	pBootstrap.args.nthread = 1
	pBootstrap.args.n       = 1000
	pBootstrap.args.stats   = 'function(x) x'
	pBootstrap.args.plot    = 'all'
	pBootstrap.args.devpars = Box(height = 2000, width = 2000, res = 300)
	pBootstrap.envs.rimport = rimport
	pBootstrap.lang         = params.Rscript.value
	pBootstrap.script       = "file:scripts/stats/pBootstrap.r"
	return pBootstrap


@procfactory
def _pPCA():
	"""
	@description:
		Perform PCA analysis using PCAtools.
		See: https://bioconductor.org/packages/release/bioc/vignettes/PCAtools/inst/doc/PCAtools.html
	@input:
		infile: The matrix to do the analysis
			- Columns are features, rows are samples
			- If not, you may use `args.params.transposed = True`
		metafile: The metadata file
			- Columns are features, rows are samples
			- rnames and cnames are required
	@output:
		outfile: The file with the components
		oudir  : The directory containing the output file and plots.
	@args:
		devpars (Box)        : The parameters for plotting device.
		inopts  (Box)        : The options to read the input files.
		na      (bool|number): How to deal with `NA` values.
			- A logistic/boolean value will remove them (use `complete.cases`)
			- Otherwise, it will be replaced by the given value.
		plots (Box): Plotting parameters supported by `PCAtools`. Set `False` to disable a plot.
			- scree: A scree plot. See `?screeplot`
				- Since `PCAtools::screeplot` returns a ggplot object, we have `args.plots.scree.ggs` to extend the plot.
			- bi: A bi-plot. See `?biplot`
			- pairs: A pairs plot. See `?pairsplot`
			- loadings: A loadings plot. See `?plotloadings`
			- eigencor: An eigencor plot. See `?eigencorplot`
		params (Box): Other parameters for `PCAtools::pca`
		npc (int|str): The number of PCs to write to the output file.
			- A fixed number of PCs; or
			- one of `horn` or `elbow` to determine the optimal number of PCs.
	"""
	return Box(
		desc   = 'Perform PCA analysis using PCAtools',
		input  = 'infile:file, metafile:file',
		output = [	'outfile:file:{{i.infile | stem | stem}}.pca/{{i.infile | stem | stem}}.pcs.txt',
					'outdir:dir:{{i.infile | stem | stem}}.pca'],
		lang = params.Rscript.value,
		args = Box(
			devpars = Box(height = 2000, width = 2000, res = 300),
			inopts  = Box(cnames = True, rnames = True),
			params  = Box(),
			na      = 0,
			plots   = Box(scree = True, bi = True, pairs = True, loadings = True, eigencor = True),
			npc     = 'elbow'
		)
	)




from pysam import VariantFile as Vcf
from pyppl import Box
from bioprocs.utils import alwaysList

infile   = {{ i.infile | quote}}
outfile  = {{ o.outfile | quote}}
rmfilter = {{ args.rmfilter | repr}}
if rmfilter:
	rmfilter = alwaysList(rmfilter)

invcf  = Vcf(infile)
outvcf = open(outfile, 'w')
outvcf.write(str(invcf.header))
for rec in invcf.fetch():
	parts   = str(rec).split('\t')
	filters = parts[6].split(';')
	if not rmfilter:
		filters = 'PASS'
	else:
		filters = ';'.join(f for f in filters if f not in rmfilter)
		filters = filters or 'PASS'
	parts[6] = filters
	outvcf.write('\t'.join(parts))
outvcf.close()





"""Rank calculations"""
from pyppl import Proc, Box
from . import params, rimport
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pRank():
	"""
	@name:
		pRank
	@description:
		Convert values to ranks.
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The output file with ranks.
	@args:
		`na`: Where to put the `NA` values.
			- `"first"` : Put `NA` first
			- `"last"`  : Put `NA` last (default)
			- `"remove"`: Remove `NA` values
			- `"keep"`  : keep `NA` values
		`tie`: How to deal with ties
			- `"average"` : Use average ranks (default)
			- `"first"`   : Use the ranks come first
			- `"last"`    : Use the ranks come last
			- `"random"`  : Use the random ranks
			- `"max"`     : Use the max ranks
			- `"min"`     : Use the min ranks
		`byrow`: Calculate ranks by row (instead of by column)? Default: `True`
		`reverse`: Take the reverse rank? Default: `True`
			- Large number gets higher rank (smaller rank index)
			- `args.na` remains the same.
		`inopts`: The input options:
			- `cnames`: Whether the input file has header. Default: `True`
			- `rnames`: Whether the input file has row names. Default: `True`
			- `delimit`: The separator of columns. Default: `\t`
	"""
	pRank              = Proc(desc = 'Convert values to ranks')
	pRank.input        = 'infile:file'
	pRank.output       = 'outfile:file:{{i.infile | fn}}.rank.txt'
	pRank.args.na      = 'last' # keep,         first,   remove
	pRank.args.tie     = 'average' # "average", "first", "last", "random", "max", "min"
	pRank.args.byrow   = True # else by column
	pRank.args.reverse = True # large number ranks higher
	pRank.args.inopts  = Box(
		cnames  = True,
		rnames  = True,
		delimit = "\t"
	)
	pRank.envs.rimport = rimport
	pRank.lang         = params.Rscript.value
	pRank.script       = "file:scripts/rank/pRank.r"
	return pRank

@procfactory
def _pRankProduct():
	"""
	@name:
		pRankProduct
	@description:
		Calculate the rank product of a set of ranks. Refer to [here](https://en.wikipedia.org/wiki/Rank_product)
	@input:
		`infile:file`: The input file
			- Format:
			```
						Case1	Case2	...
			Feature1	8.2  	10.1 	...
			Feature2	2.3  	8.0  	...
			...
			```
			- Or instead of values, you can also have ranks in the input file:
			```
						Rank1	Rank2	...
			Feature1	2    	1    	...
			Feature2	3    	2    	...
			...
			```
	@output:
		`outfile:file`: The output file with original ranks, rank products and p-value if required
	@args:
		`informat`: The input format of the values. Whether they are real values (value) or ranks (rank). Default: value
			- Records will be ordered descendingly by value (Larger value has higher rank (lower rank index)).
		`pval`:     Whether to calculate the p-value or not. Default: True
		`plot`:     Number of rows to plot. Default: 0 (Don't plot)
		`cex`:      Font size for plotting. Default: 0.9
		`cnheight`: Colname height. Default: 80
		`rnwidth`:  Rowname width. Default: 50
		`devpars`:  device parameters for the plot. Default: `Box(res=300, width=2000, height=2000)`
		`inopts`:   Options for reading the input file. Default: `Box(cnames=True, rnames=True, delimit="\t")`
	"""
	pRankProduct               = Proc(desc = 'Calculate the rank product of a set of ranks.')
	pRankProduct.input         = "infile:file"
	pRankProduct.output        = "outdir:dir:{{i.infile | fn}}.rp"
	pRankProduct.envs.rimport  = rimport
	pRankProduct.args.informat = "value"
	pRankProduct.args.pval     = True
	pRankProduct.args.cex      = .9
	pRankProduct.args.cnheight = 80
	pRankProduct.args.rnwidth  = 50
	pRankProduct.args.inopts   = Box(cnames  = True, rnames  = True, delimit = '\t')
	pRankProduct.args.devpars  = Box(res = 300, width = 2000, height = 2000)
	pRankProduct.lang          = params.Rscript.value
	pRankProduct.script        = "file:scripts/rank/pRankProduct.r"
	return pRankProduct




from os import path
from pyppl import Box
from bioprocs.utils import runcmd, cmdargs

params = {}
region = {{i.region | quote}}
outfile = {{o.outfile | quote}}

if path.isfile(region):
	if path.getsize(region) > 0:
		params['R'] = {{i.region | quote}}
		params.update({{args.params}})
		cmd = '{{args.tabix}} %s "{{i.infile}}" > "{{o.outfile}}"' % cmdargs(params, equal=' ')
		runcmd(cmd)
	else:
		open(outfile, 'w').close()
else:
	if region:
		params.update({{args.params}})
		cmd = '{{args.tabix}} %s "{{i.infile}}" {{i.region}} > "{{o.outfile}}"' % cmdargs(params, equal=' ')
		runcmd(cmd)
	else:
		open(outfile, 'w').close()



"""Gene set enrichment analysis"""
from os import path
from pyppl import Proc, Box
from . import params
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pGMT2Mat():
	"""
	@name:
		pGMT2Mat
	@description:
		Convert a GMT file to a matrix.
		Rownames of GMT file will be the column names of output matrix.
	@input:
		`infile:file`: The input file in GMT format.
	@output:
		`outfile:file`: output matrix file
	"""
	pGMT2Mat        = Proc(desc = 'Convert a GMT file to a matrix.')
	pGMT2Mat.input  = "infile:file"
	pGMT2Mat.output = "outfile:file:{{i.infile | fn}}.gmat"
	pGMT2Mat.lang   = params.python.value
	pGMT2Mat.script = "file:scripts/gsea/pGMT2Mat.py"
	return pGMT2Mat

@procfactory
def _pExprMat2GCT():
	"""
	@name:
		pExprMat2GCT
	@description:
		Convert expression matrix to GCT file.
		Refer to http://software.broadinstitute.org/cancer/software/genepattern/file-formats-guide#GCT for file format
	@input:
		`expfile:file`: the input expression matrix file. Samples as columns, genes as rows.
	@output:
		`outfile:file`: the gct file
	"""
	pExprMat2GCT        = Proc(desc = 'Convert expression matrix to GCT file.')
	pExprMat2GCT.input  = 'expfile:file'
	pExprMat2GCT.output = 'outfile:file:{{ i.expfile | fn }}.gct'
	pExprMat2GCT.lang   = params.python.value
	pExprMat2GCT.script = "file:scripts/gsea/pExprMat2GCT.py"
	return pExprMat2GCT

@procfactory
def _pSampleinfo2CLS():
	"""
	@name:
		pSampleinfo2CLS
	@description:
		Convert sample infomation to cls file.
		Refer to http://software.broadinstitute.org/cancer/software/genepattern/file-formats-guide#CLS for file format
		NOTE that the order of samples must be the same as in GMT file in further analysis.
	@input:
		`sifile:file`: the sample information file.
			- Headers are: [Sample, ]Patient, Group, Batch
			- Rows are samples
	@output:
		`outfile:file`: the cls file
	"""
	pSampleinfo2CLS                    = Proc(desc = 'Convert sample infomation to cls file.')
	pSampleinfo2CLS.input              = 'sifile:file'
	pSampleinfo2CLS.output             = 'outfile:file:{{ i.sifile | fn }}.cls'
	#pSampleinfo2CLS.envs.txtSampleinfo = txt.sampleinfo.py
	pSampleinfo2CLS.lang               = params.python.value
	pSampleinfo2CLS.script             = "file:scripts/gsea/pSampleinfo2CLS.py"
	return pSampleinfo2CLS

@procfactory
def _pSSGSEA():
	"""
	@name:
		pSSGSEA
	@description:
		Single sample GSEA
		Refer to http://software.broadinstitute.org/cancer/software/genepattern/file-formats-guide#GCT for GCT file format
		Refer to http://software.broadinstitute.org/cancer/software/genepattern/file-formats-guide#GMT for GMT file format
	@input:
		`gctfile:file`: the expression file
		`gmtfile:file`: the gmtfile for gene sets
	@output:
		`outdir:file`: the output directory
		- `report.txt`: the enrichment report for each Gene set.
		- `RES_<GeneSet>.png`: the running ES plot for <GeneSet>
		- `normP_<GeneSet>.png`: the norminal P value plot for <GeneSet>
	@args:
		`weightexp`: Exponential weight employed in calculation of enrichment scores. Default: 0.75
		`nperm`:     Number of permutations. Default: 10000
	"""
	pSSGSEA = Proc (desc = 'Do single-sample GSEA.')
	pSSGSEA.input          = "gctfile:file, gmtfile:file"
	pSSGSEA.output         = "outdir:file:{{i.gctfile | fn}}-{{i.gmtfile | fn}}-ssGSEA"
	pSSGSEA.args.weightexp = 1
	pSSGSEA.args.nperm     = 1000
	pSSGSEA.args.seed      = -1
	pSSGSEA.lang           = params.Rscript.value
	pSSGSEA.script         = "file:scripts/gsea/pSSGSEA.r"
	return pSSGSEA

@procfactory
def _pGSEA():
	"""
	@name:
		pGSEA
	@description:
		GSEA
		Refer to http://software.broadinstitute.org/cancer/software/genepattern/file-formats-guide#GCT for GCT file format
		Refer to http://software.broadinstitute.org/cancer/software/genepattern/file-formats-guide#GMT for GMT file format
		Refer to http://software.broadinstitute.org/cancer/software/genepattern/file-formats-guide#CLS for CLS file format
	@input:
		`gctfile:file`: the expression file
		`clsfile:file`: the class file
		`gmtfile:file`: the gmtfile for gene sets
	@output:
		`outdir:file`: the output directory
	@args:
		`weightexp`: Exponential weight employed in calculation of enrichment scores. Default: 0.75
		`nperm`:     Number of permutations. Default: 1000
	"""
	pGSEA                = Proc (desc = 'Do GSEA.')
	pGSEA.input          = "gctfile:file, clsfile:file, gmtfile:file"
	pGSEA.output         = "outdir:dir:{{i.gctfile | fn}}-{{i.gmtfile | fn}}.GSEA"
	pGSEA.args.weightexp = 1
	pGSEA.args.nperm     = 1000
	pGSEA.args.nthread   = 1
	pGSEA.args.seed      = -1
	pGSEA.lang           = params.Rscript.value
	pGSEA.script         = "file:scripts/gsea/pGSEA.r"
	return pGSEA

@procfactory
def _pEnrichr():
	"""
	@input:
		`infile:file`: The gene list, each per line
	@output:
		`outdir:dir`:  The output directory, containing the tables and figures.
	@args:
		`top`:     Top N pathways used to plot
		`genecol`: The columns index containing the genes
		`inopts`:  The input options.
			- `delimit`: The delimit of input file.
			- `skip`:    Skip first N lines.
			- `comment`: Line comment mark.
			- Other parameters fit `bioprocs.utils.tsvio.TsvReader`
		`libs`:  The databases to do enrichment against.
			- A full list can be found here: http://amp.pharm.mssm.edu/Enrichr/#stats
			- Multiple dbs separated by comma (,)
		`plot`: Whether to plot the result.
		`devpars`: Parameters for png.
		include: A lambda function to include the records(genes)
			- argument is `bioprocs.utils.tsvio2.TsvRecord`
	"""
	return Box(
		desc = 'Gene set enrichment analysis using Enrichr APIs',
		input = 'infile:file',
		output = 'outdir:dir:{{i.infile | stem}}.enrichr',
		lang = params.python.value,
		errhow = 'retry',
		args = Box(
			inopts   = Box(delimit = '\t', skip = 0, comment = '#'),
			top      = 20,
			cutoff   = 1,
			genecol  = 0,
			include  = None,
			nthread  = 1,
			Rscript  = params.Rscript.value,
			pathview = Box(), # Box(fccol = 2)
			libs     = "KEGG_2019_Human",
			devpars  = Box(res = 300, width = 2000, height = 2000),
			plot     = True
		)
	)

@procfactory
def _pGene2Pathway():
	"""
	@name:
		pGene2Pathway
	@description:
		Find pathways where genes are present
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The output file, Default: `{{i.infile | fn}}-pw{{i.infile | ext}}`
	@args:
		`inopts`: Reading options for input file, Default: `Box(cnames = True)`
		`genecol`: Index or name of the gene column, Default: `0`
		`libs`: Libraries of the pathways, Default: `KEGG_2019_Human`
	"""
	pGene2Pathway              = Proc(desc = 'Find pathways that genes are present.')
	pGene2Pathway.input        = 'infile:file'
	pGene2Pathway.output       = 'outfile:file:{{i.infile | fn}}-pw{{i.infile | ext}}'
	pGene2Pathway.args.inopts  = Box(cnames = True)
	pGene2Pathway.args.genecol = 0
	pGene2Pathway.args.libs    = "KEGG_2019_Human"
	pGene2Pathway.lang         = params.python.value
	pGene2Pathway.script       = "file:scripts/gsea/pGene2Pathway.py"
	return pGene2Pathway

@procfactory
def _pTargetEnrichr():
	"""
	@name:
		pTargetEnrichr
	@description:
		Use APIs from http://amp.pharm.mssm.edu/Enrichr/help#api&q=1 to analyze a gene list
	@input:
		`infile:file`: The target genes with regulators
			- Format (RegulatorStatus and TargetStatus are optional):
			```
			Regulator	Target	RegulatorStatus	TargetStatus	Relation
			has-mir-22	Gene	+	+	+
			```
	@output:
		`outdir:dir`:  The output directory, containing the tables and figures.
	@args:
		`dbs`       : The databases to do enrichment against. Default: KEGG_2016
		- A full list can be found here: http://amp.pharm.mssm.edu/Enrichr/#stats
		- Multiple dbs separated by comma (,)
		`rmtags`    : Remove pathway tags in the plot. Default: True
		- For example: change "Lysine degradation_Homo sapiens_hsa00310" to "Lysine degradation".
		`enrplot`   : Whether to plot the result. Default: True
		`enrn`      : Top N pathways used to plot. Default: 10
		`netplot`   : Whether to plot the network. Default: True
		`netn`      : Top N pathways used to plot the network. Default: 5
			- Must <= `enrn`. If `netn` >= `enrn`, `netn` = `enrn`
		`title`     : The title for the plot. Default: "Gene enrichment: {db}"
	@requires:
		[`python-mygene`](https://pypi.python.org/pypi/mygene/3.0.0)
		[`graphviz`](https://pypi.python.org/pypi/graphviz)
	"""
	pTargetEnrichr               = Proc(desc = 'Do gene set enrichment analysis for target genes.')
	pTargetEnrichr.input         = "infile:file"
	pTargetEnrichr.output        = "outdir:dir:{{i.infile | fn}}.tenrichr"
	pTargetEnrichr.lang          = params.python.value
	pTargetEnrichr.args.inopts   = Box(delimit = '\t', skip = 0, comment = '#', ftype = 'head')
	pTargetEnrichr.args.genecol  = "COL2"
	pTargetEnrichr.args.dbs      = "KEGG_2016"
	pTargetEnrichr.args.norm     = False
	pTargetEnrichr.args.rmtags   = True
	pTargetEnrichr.args.enrplot  = True
	pTargetEnrichr.args.enrn     = 10
	pTargetEnrichr.args.netplot  = True
	pTargetEnrichr.args.netn     = 5
	pTargetEnrichr.args.title    = "Gene enrichment: {db}"
	pTargetEnrichr.args.cachedir = params.cachedir.value
	#pTargetEnrichr.envs.genenorm = genenorm.py
	pTargetEnrichr.errhow        = 'retry'
	pTargetEnrichr.script        = "file:scripts/gsea/pTargetEnrichr.py"
	return pTargetEnrichr




from pyppl import Box
from bioprocs.utils.tsvio2 import TsvReader

infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}
font    = {{args.font | repr}}
section = {{args.section | repr}}
{% case True %}
{% when isinstance(args.title, (tuple, list)) %}
title = {{args.title | lambda a, render = render: (render(a[0]), a[1])}}
{% when args.title | bool %}
title = {{args.title | lambda a, render = render: (render(a), 1)}}
{% else %}
title = {{args.title | repr}}
{% endcase %}
caption = {{args.caption | render | repr}}
style   = {{args.style | repr}}
inopts  = {{args.inopts | repr}}
acode   = {{args.acode | repr}}
bcode   = {{args.bcode | repr}}

# fix parameters
if section:
	section['type']   = section.get('type', 'NEW_PAGE').upper()
	section['orient'] = section.get('orient', 'PORTRAIT').upper()
	section['margin'] = section.get('margin', [36] * 4)
	if isinstance(section['margin'], int):
		section['margin'] = [section['margin']] * 4
	elif len(section['margin']) == 2:
		section['margin'].extend(section['margin'])
	elif len(section['margin']) == 3:
		section['margin'].append(section['margin'][1])

if not isinstance(bcode, (tuple, list)):
	bcode = [bcode]
bcode = "\n".join(bcode)

if not isinstance(acode, (tuple, list)):
	acode = [acode]
acode = "\n".join(acode)

if caption:
	caption = caption.split('.', 1) if caption.startswith('Table ') and '.' in caption else ['', caption]

# region template
template = """
# Set up section
{section_code}

# Add heading
{heading_code}

# Interpolation code before table
{bcode_code}

# Add caption
{caption_code}

# Insert the table
{table_code}

# Interpolation code after table
{acode_code}
"""
# endregion

# region section_code
section_code = """
sec             = doc.add_section()
sec.type        = WD_SECTION.{sectype}
sec.orientation = WD_ORIENT.{secorient}

_new_width, _new_height = sec.page_height, sec.page_width
if sec.orientation != doc.sections[-2].orientation:
	sec.page_width, sec.page_height = _new_width, _new_height
else:
	sec.page_width, sec.page_height = _new_height, _new_width
sec.top_margin, sec.right_margin, sec.bottom_margin, sec.left_margin = (Pt(x) for x in {secmargin!r})
""".format(
	sectype   = section['type'],
	secorient = section['orient'],
	secmargin = section['margin']
) if section else ""
# endregion

# region heading_code
heading_code = """
doc.add_heading('''{}''', {})
""".format(*title) if title else ""
# endregion

bcode_code = bcode

caption_code = """
_cap_para = doc.add_paragraph()
{alignment}
_cap_para.add_run({caption0!r}).font.bold = True
_cap_para.add_run({caption1!r})
""".format(
	alignment = "_cap_para.alignment = WD_ALIGN_PARAGRAPH.{}\n".format(align.upper()) if align else '',
	caption0   = caption[0],
	caption1   = caption[1]
) if caption else ''

inopts['cname0'] = inopts.get('cname0', '')
reader = TsvReader(infile, **inopts)
table_code = """
cnames = {cnames!r}
data   = {data!r}
if cnames:
	data.insert(0, cnames)
table = doc.add_table(rows = len(data), cols = len(cnames))
table.style = {style!r}
for i, row in enumerate(data):
	for j, r in enumerate(row):
		table.rows[i].cells[j].text = r
		table.rows[i].cells[j].paragraphs[0].runs[0].font.name = {family!r}
		table.rows[i].cells[j].paragraphs[0].runs[0].font.size = Pt({size!r})
""".format(
	cnames = reader.cnames,
	data   = [r.values() for r in reader],
	family = font['family'],
	size   = font['size'],
	style  = style
)

acode_code = acode

with open(outfile, 'w') as f:
	f.write(template.format(
		section_code = section_code,
		heading_code = heading_code,
		bcode_code   = bcode_code,
		caption_code = caption_code,
		table_code   = table_code,
		acode_code   = acode_code
	))




from os import path
from collections import OrderedDict
from pyppl import Box
from pyppl.utils import alwaysList
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter

infile  = {{i.infile | quote}}
colfile = {{i.colfile | quote}}
outfile = {{o.outfile | quote}}
inopts  = {{args.inopts | repr}}
cols    = {{args.cols | repr}}
keep    = {{args.keep | repr}}

if path.isfile(colfile):
	cols = TsvReader(colfile, cnames = False).dump(0)
elif colfile:
	cols = alwaysList(colfile)
elif path.isfile(str(cols)):
	cols = TsvReader(cols, cnames = False).dump(0)
elif cols:
	cols = alwaysList(cols)
else:
	raise ValueError('Columns not provided.')
if not isinstance(cols[0], int) and cols[0].isdigit():
	cols = [int(c) for c in cols]

reader = TsvReader(infile, **inopts)
writer = TsvWriter(outfile, delimit = inopts.get('delimit', "\t"))
if reader.cnames and not isinstance(cols[0], int):
	cols = [reader.cnames.index(c) for c in cols if c in reader.cnames]
elif not reader.cnames and not isinstance(cols[0], int):
	raise ValueError("Input file doesn't have column names")
elif min(cols) < -len(reader.cnames) or (reader.cnames and max(cols) >= len(reader.cnames)):
	raise IndexError("Provided columns beyond input file range.")

if reader.cnames:
	ncol = len(reader.cnames)
else:
	ncol = len(reader.next())
	reader.rewind()

cols = [ncol + c if c < 0 else c for c in cols]
if not keep:
	cols = [c for c in range(ncol) if c not in cols]

if reader.cnames:
	writer.cnames = [reader.cnames[c] for c in cols]
	writer.writeHead()

for r in reader:
	rec = [r[c] for c in cols]
	writer.write(rec)
writer.close()



from pyppl import Box
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter

infile  = {{ i.infile | quote}}
outfile = {{o.outfile | quote}}
out     = {{args.out | quote}}
outopts = {{args.outopts |repr}}

CNAMES = Box(
	tumor  = ['TUMOR'],
	normal = ['NORMAL'],
	both   = ['TUMOR', 'NORMAL']
)


reader = TsvReader(infile)
writer = TsvWriter(outfile)
cnames = outopts.get('cnames')
if cnames is True:
	cnames = CNAMES[out]
if cnames:
	writer.cnames = cnames
	writer.writeHead()

prev_tumor = prev_normal = None
for r in reader:
	tumor, normal = r.Tumor_Sample_Barcode, r.Matched_Norm_Sample_Barcode
	if prev_tumor == tumor and prev_normal == normal:
		continue
	if out == 'tumor':
		writer.write([tumor])
	elif out == 'normal':
		writer.write([normal])
	else:
		writer.write([tumor, normal])
	prev_tumor = tumor
	prev_normal = normal



from os import path
from pyppl import Box
from bioprocs.utils import shell
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter

fq1     = {{ i.fqfile1 | quote}}
fq2     = {{ i.fqfile2 | quote}}
outfile = {{ o.outfile | quote}}
outdir  = {{ o.outdir | quote}}
params = {{ args.params | repr}}
idxfile = {{ args.idxfile | quote}}
kallisto = {{ args.kallisto | quote}}
nthread = {{ args.nthread | repr}}

shell.TOOLS.kallisto = kallisto
params.i = idxfile
params.o = outdir
params.t = nthread
params._ = [fq1, fq2]

kallisto = shell.Shell(subcmd = True).kallisto
kallisto.quant(**params).run()

imfile        = path.join(outdir, 'abundance.tsv')
reader        = TsvReader(imfile)
writer        = TsvWriter(outfile)
writer.cnames = ['target_id', 'est_counts']
writer.writeHead()

for r in reader:
	r.target_id = r.target_id.split('::')[0]
	try:
		r.est_counts = int(round(float(r.est_counts)))
	except TypeError:
		r.est_counts = 0
	writer.write(r)
writer.close()



#!/usr/bin/env python
"""Plot genomic elements."""
import json, sys
from os import path
from pyppl import PyPPL, utils, Box
from bioprocs import params

params._prefix = '-'

params.ideo       = 'True'
params.ideo.desc  = 'Show ideogram track?'
params.genes      = params.refgene.value
params.genes.desc = 'Show gene track?'
params.axis       = True
params.axis.desc  = 'Show axis?'

params.genome.show     = True
params.outdir.required = True
params.outdir.desc     = 'Output directory.'

params.region.required = True
params.region.desc     = 'The region to plot. E.g. chr1:1000000-1200000'
params.tracks.required = True
params.tracks.type     = list
params.tracks.desc     = 'The track types. Could be data, anno, interaction or ucsc, or multiple of them.'
params.names.required  = True
params.names.type      = list
params.names.desc      = 'The corresponding names of the tracks.'
params.inputs.required = True
params.inputs.type     = list
params.inputs.desc     = 'The input of the tracks.\n"<ucscTrack>:<gvizTrack>" for ucsc track;\n"<infile>:<intype>" for interaction tracks;\nfiles for other tracks.'
params.params.type     = list
params.params.desc     = 'The params for each track'
params.plotparams.desc = 'The params for pGenomePlot.'
params.devpars         = '{"res": 300, "height": 300, "width": 2000}'
params.devpars.desc    = 'The device parameters for plotting.'
params.splitlen        = 5000000
params.splitlen.desc   = 'Split the plot into 2 if the region is longer then splitlen.'
params.forks           = 2
params.forks.desc      = 'Number of cores used to plot if split.'

# highlist
params.highlights      = []
params.highlights.desc = 'The highlight regions in format of "start-end"'

def main():
	opts = params._parse(dict_wrapper = Box)
	from bioprocs.genomeplot import pInteractionTrack, pAnnoTrack, pDataTrack, \
		pUcscTrack, pGenomePlot

	chrom, startend  = opts.region.split(':')
	start, end       = startend.split('-')
	start      = int(start)
	end        = int(end)
	trackProcs = []
	#uuid       = utils.uid(str(sys.argv))
	for i, tt in enumerate(opts.tracks):
		if tt == 'data':
			datatrackproc = pDataTrack.copy(tag = opts.names[i])
			datatrackproc.input = (opts.names[i], opts.inputs[i], chrom)
			if opts.params:
				datatrackproc.args.opts.update(json.loads(opts.params[i]))
			trackProcs.append(datatrackproc)
		elif tt == 'anno':
			annotrackproc = pAnnoTrack.copy(tag = opts.names[i])
			annotrackproc.input = (opts.names[i], opts.inputs[i], chrom)
			if opts.params:
				annotrackproc.args.opts.update(json.loads(opts.params[i]))
			trackProcs.append(annotrackproc)
		elif tt == 'ucsc':
			ucsctrackproc = pUcscTrack.copy(tag = opts.names[i])
			ucsctrack, gviztrack = opts.inputs[i].split(':')
			ucsctrackproc.input = (opts.names[i], ucsctrack, gviztrack, opts.region)
			if opts.params:
				ucsctrackproc.args.opts.update(json.loads(opts.params[i]))
			trackProcs.append(ucsctrackproc)
		else:
			intertrackproc = pInteractionTrack.copy(tag = opts.names[i])
			infile, intype = opts.inputs[i].split(':')
			intertrackproc.input = (opts.names[i], infile, opts.region)
			intertrackproc.args.intype = intype
			if opts.params:
				intertrackproc.args.opts.update(json.loads(opts.params[i]))
			trackProcs.append(intertrackproc)

	if end - start > opts.splitlen:
		pGenomePlot.depends        = trackProcs
		#pGenomePlot.tag            = uuid
		pGenomePlot.forks          = opts.forks
		pGenomePlot.exdir          = opts.outdir
		pGenomePlot.args.ideoTrack = opts.ideo
		pGenomePlot.args.axisTrack = opts.axis
		pGenomePlot.args.geneTrack = opts.genes
		if opts.devpars:
			pGenomePlot.args.devpars.update(json.loads(opts.devpars))
		if opts.plotparams:
			pGenomePlot.args.opts.update(json.loads(opts.plotparams))
		if len(opts.highlights) == 2 and '-' not in opts.highlist[0]:
			h1 = opts.highlights[0]
			h2 = opts.highlights[1]
		else:
			h1 = ';'.join(opts.highlights)
			h2 = h1
		pGenomePlot.input   = lambda *chs: [
			([ch.get() for ch in chs], "%s:%s-%s" % (chrom, start, start + 10000), h1),
			([ch.get() for ch in chs], "%s:%s-%s" % (chrom, end - 100000, end), h2),
		]

	else:
		pGenomePlot.depends        = trackProcs
		#pGenomePlot.tag            = uuid
		pGenomePlot.exdir          = opts.outdir
		pGenomePlot.args.ideoTrack = opts.ideo
		pGenomePlot.args.axisTrack = opts.axis
		pGenomePlot.args.geneTrack = opts.genes
		if opts.devpars:
			pGenomePlot.args.devpars.update(json.loads(opts.devpars))
		if opts.plotparams:
			pGenomePlot.args.opts.update(json.loads(opts.plotparams))
		pGenomePlot.input   = lambda *chs: [([ch.get() for ch in chs], opts.region, ';'.join(opts.highlights))]

	config = {'proc': {'file': None}}
	PyPPL(config).start(trackProcs).run()

if __name__ == "__main__":
	main()



"""A set of algorithms or models"""
from pyppl import Proc, Box
from . import params, rimport
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pRWR():
	"""
	@name:
		pRWR
	@description:
		Do random walk with restart (RWR)
	@input:
		`Wfile:file`: The adjecent matrix
		`Efile:file`: The start vector
	@output:
		`outfile:file`: The output of final probabilities
	@args:
		`c`:       The restart probability. Default: 0.1
		`eps`:     The convergent cutoff || R(i+1) - R(i) ||. Default: 1e-5
		`niter`:   Max iterations to stop. Default: 10000
		`normW`:   Weather to normalize W or not, default True.
			- Laplacian normalization is used (more to add).
		`normE`:   Weather to normalize E or not, default True.
			- E will be normalized as: E = E/sum(E)
	@requires:
		if normW = True, R package `NetPreProc` is required.
	"""
	pRWR            = Proc (desc = 'Do random walk with restart (RWR).')
	pRWR.input      = "Wfile:file, Efile:file"
	pRWR.output     = "outfile:file:{{i.Wfile | fn2}}.rwr.txt"
	pRWR.args.c     = 0.1
	pRWR.args.eps   = 1e-5
	pRWR.args.niter = 10000
	pRWR.args.normW = True
	pRWR.args.normE = True
	pRWR.lang       = params.Rscript.value
	pRWR.script     = "file:scripts/algorithm/pRWR.r"
	return pRWR

@procfactory
def _pAR():
	"""
	@name:
		pAR
	@description:
		Affinity Regression.
		Ref: https://www.nature.com/articles/nbt.3343
		```
				b           c        d          d
			_________    _______    ____       ____
			|       |    |  W  |    |  |       |  |
		a |   D   |  b |_____|  c |Pt|  =  a |Y |   <=>
			|_______|               |__|       |  |
											|__|
		kronecker(P, YtD)*vec(W) = vec(YtY)             <=>
		X*vec(W) = vec(YtY)
		WPt:
			c        d              d
			_______    ____          _____
			|  W  |    |  |          |   |
		b |_____|  c |Pt|  --->  b |___|
					|__|
		YtDW:
		WtDtY:
			b           a        d               d
		_______    _________   ____           _____
		|  Wt |    |       |   |  |           |   |
		c |_____|  b |   Dt  | a |Y |    ---> c |___|
					|_______|   |  |
								|__|
		```
	@input:
		`D:file` : The D matrix
		`Pt:file`: The Pt matrix
		`Y:file`:  The Y matrix
			- All input files could be gzipped
	@output:
		`W:file`:  The interaction matrix
		`outdir:dir`: The output directory
	@args:
		`seed`:  The seed for sampling the training set.
		`tfrac`: The fraction of samples used for training.
	"""
	pAR            = Proc(desc =  'Affinity Regression.')
	pAR.input      = 'D:file, Pt:file, Y:file'
	pAR.output     = [
		'W:file:{{i.D | fn}}-{{i.Pt | fn}}-{{i.Y | fn}}.AR/W.txt',
		'outdir:dir:{{i.D | fn}}-{{i.Pt | fn}}-{{i.Y | fn}}.AR'
	]
	pAR.args.seed    = None
	pAR.args.tfrac   = .5
	pAR.args.inopts  = Box(cnames = True, rnames = True)
	pAR.args.svdP    = 0
	pAR.args.predY   = True
	pAR.args.WPt     = True
	pAR.args.WtDtY   = True
	pAR.args.nfold   = 3
	pAR.args.nthread = 1
	pAR.args.method  = 'glmnet' # admm
	pAR.envs.rimport = rimport
	pAR.lang         = params.Rscript.value
	pAR.script       = "file:scripts/algorithm/pAR.r"
	return pAR

@procfactory
def _pColoc():
	"""
	@name:
		pColoc
	@description:
		Bayes Factor colocalisation analyses using R `coloc` package.
		`coloc` package can accept multiple formats of input. Here we adopt the one using pvalues.
		`coloc.abf(dataset1=list(pvalues=p1,N=nrow(X1),type="quant"), dataset2=list(pvalues=p2,N=nrow(X2),type="quant"), MAF=maf)`
	@input:
		`infile:file`: The input file including the MAF, pvalues of 1st and 2nd phenotypes
			- The first 6 columns are in BED6 format.
			- 7th : MAF
			- 8th : Pvalues for the 1st phenotype
			- 9th : Pvalues for the 2nd phenotype
			- This file could have a header with the names for phenotypes
			- Snps have to be on the same chromosome, and sorted by positions.
	@output:
		`outfile:file`: The output file including:
			- # snps, PP.H0.abf, PP.H1.abf, PP.H2.abf, PP.H3.abf and PP.H4.abf
		`outdir:dir`  : The output directory containing the output file and plots.
	@args:
		`plot`: Do manhattan plot? Default: `True`
	"""
	pColoc        = Proc(desc = "Bayes Factor colocalisation analyses using R `coloc` package.")
	pColoc.input  = 'infile:file'
	pColoc.output = [
		'outfile:file:{{i.infile | fn2}}.coloc/{{i.infile | fn2}}.coloc.txt',
		'outdir:dir:{{i.infile | fn2}}.coloc'
	]
	pColoc.args.inopts  = Box(cnames = True, rnames = False)
	pColoc.args.plot    = True
	pColoc.args.ggs     = Box()
	pColoc.args.params  = Box()
	pColoc.args.devpars = Box(res = 300, height = 2000, width = 2000)
	pColoc.args.hifile  = ''
	pColoc.args.hilabel = True
	pColoc.envs.rimport = rimport
	pColoc.lang         = params.Rscript.value
	pColoc.script       = "file:scripts/algorithm/pColoc.r"
	return pColoc





from pyppl import Box
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter

infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}

# Keep the comments
with open(infile) as fin, open(outfile, 'w') as fout:
	for line in fin:
		if line.startswith('#'):
			fout.write(line)
		else:
			break

reader = TsvReader(infile)
writer = TsvWriter(outfile, append = True)
writer.cnames = reader.cnames
writer.writeHead()
for r in reader:
	if not r.Chromosome.startswith('chr'):
		r.Chromosome = 'chr' + r.Chromosome
	writer.write(r)
reader.close()
writer.close()
